# coding: utf-8
# Distributed under the terms of the MIT License.

""" This module implements the Electrode class used for calculating
relevant electrode properties from phase diagrams.

"""

__all__ = ["Electrode", "VoltageProfile"]
__author__ = "Matthew Evans"
__maintainer__ = "Matthew Evans"


from .electrode import Electrode, VoltageProfile



# coding: utf-8
# Distributed under the terms of the MIT License.

""" This submodule contains functions to plot chemical shifts
and chemical shieldings from NMR calculations.

"""

from typing import List, Optional, Dict, Union, Tuple

import numpy as np
from matador.crystal import Crystal
from matador.plotting.plotting import plotting_function
from matador.fingerprints import Fingerprint

__all__ = ["plot_magres"]


@plotting_function
def plot_magres(
    magres: Union[List[Crystal], Crystal],
    species: str,
    magres_key: str = "chemical_shielding_iso",
    xlabel: str = None,
    broadening_width: float = 1,
    text_offset: float = 0.2,
    ax=None,
    figsize: Tuple[float, float] = None,
    show: bool = False,
    savefig: Optional[str] = None,
    signal_labels: Optional[Union[str, List[str]]] = None,
    signal_limits: Tuple[float, float] = None,
    line_kwargs: Optional[Union[Dict, List[Dict]]] = None,
    invert_xaxis: bool = True,
    species_label: Optional[str] = None,
    label_fontsize: int = None,
):
    """Plot magnetic resonance data for a set of crystals.

    Parameters:
        magres (Union[Crystal, List[Crystal]]): list of :class:`Crystal` containing
            magres data.
        species (str | Tuple[str, int]): the species to plot the shifts of, either as
            a simple element symbol, or an element and isotope tuple used for labelling,
            e.g., "P" or ("P", 31).

    Keyword arguments:
        ax (matplotlib.axes.Axes): an existing axis on which to plot.
        magres_key (str): the data key for which the magres site data is stored under.
        show (bool): whether to show plot in an X window.
        figsize (Tuple[float]): overrides the default size for the matplotlib figure.
        broadening_width (float): the Lorentzian width to apply to the shifts.
        xlabel (str): a custom label for the x-axis.
        savefig (str): filename to use to save the plot.
        signal_labels (list): optional list of labels for the curves in
            the magres list.
        signal_limits (Tuple[float]): values at which to clip the magres signals. Defaults
            to the maximum and minimum shifts across all passed structures.
        line_kwargs (list or dict): parameters to pass to the curve plotter,
            if a list then the line kwargs will be passed to each line individually.
        label_fontsize (int): font size override for the labels.

    """
    import matplotlib.pyplot as plt

    if not isinstance(magres, list):
        magres = [magres]
    if signal_labels is not None and not isinstance(signal_labels, list):
        signal_labels = [signal_labels]
    if line_kwargs is not None and not isinstance(line_kwargs, list):
        line_kwargs = len(magres) * [line_kwargs]

    if figsize is None:
        _user_default_figsize = plt.rcParams.get("figure.figsize", (8, 6))
        height = len(magres) * max(0.5, _user_default_figsize[1] / 1.5 / len(magres))
        figsize = (_user_default_figsize[0], height)

    if ax is None:
        fig = plt.figure(figsize=figsize)
        ax = fig.add_subplot(111)

    if species is None:
        raise RuntimeError("You must provide the species of interest plotting.")
    if isinstance(species, tuple):
        species_label = f"$^{{ {species[1]} }}${species[0]}"
        species = species[0]
    else:
        species_label = species

    if signal_labels is not None and len(signal_labels) != len(magres):
        raise RuntimeError(
            "Wrong number of labels passed for number of magres: {} vs {}".format(
                len(signal_labels), len(magres)
            )
        )

    _magres = []

    if signal_limits is not None:
        min_shielding, max_shielding = sorted(signal_limits)
        if signal_limits[0] > signal_limits[1]:
            ax.invert_xaxis()
    else:
        min_shielding, max_shielding = (1e20, -1e20)

    for ind, doc in enumerate(magres):

        if isinstance(doc, dict):
            _doc = Crystal(doc)
        else:
            _doc = doc

        _magres.append(_doc)

        relevant_sites = [atom for atom in _doc if atom.species == species]
        if relevant_sites:
            shielding = [atom[magres_key] for atom in relevant_sites]
            if signal_limits is None:
                min_shielding = min(np.min(shielding), min_shielding)
                max_shielding = max(np.max(shielding), max_shielding)

    if min_shielding > 1e19 and max_shielding < -1e19:
        raise RuntimeError(
            f"No sites of {species} found in any of the passed crystals."
        )

    _buffer = 0.2 * np.abs(min_shielding - max_shielding)
    s_space = np.linspace(min_shielding - _buffer, max_shielding + _buffer, num=1000)

    _padded_colours = list(plt.rcParams["axes.prop_cycle"].by_key()["color"])
    _padded_colours = (1 + (len(magres) // len(_padded_colours))) * _padded_colours

    if line_kwargs is not None and len(line_kwargs) != len(magres):
        raise RuntimeError(
            "Wrong number of line kwargs passed for number of magres: {} vs {}".format(
                len(line_kwargs), len(magres)
            )
        )

    for ind, doc in enumerate(_magres):
        if signal_labels is None:
            stoich_label = doc.formula_tex
        else:
            stoich_label = None

        _label = stoich_label
        if signal_labels is not None and len(signal_labels) > ind:
            _label = signal_labels[ind]

        _line_kwargs = {"c": _padded_colours[ind]}
        if line_kwargs is not None:
            _line_kwargs.update(line_kwargs[ind])

        relevant_sites = [site for site in doc if site.species == species]
        if not relevant_sites:
            print(
                f"No sites of {species} found in {doc.root_source}, signal will be empty."
            )
            signal = np.zeros_like(s_space)

        else:
            shifts = [site[magres_key] for site in relevant_sites]

            hist, bins = np.histogram(shifts, bins=s_space)

            if broadening_width > 0:
                signal = Fingerprint._broadening_unrolled(
                    hist, s_space, broadening_width, broadening_type="lorentzian"
                )
            else:
                signal = np.array(hist, dtype=np.float64)
                bin_centres = s_space[:-1] + (s_space[1] - s_space[0]) / 2
                s_space = bin_centres

            if np.max(signal) > 1e-10:
                signal /= np.max(signal)
            else:
                signal *= 0

        ax.plot(s_space, signal + (ind * 1.1), **_line_kwargs)

        if _label is not None:
            ax.text(
                0.95,
                (ind * 1.1) + text_offset,
                _label,
                transform=ax.get_yaxis_transform(),
                horizontalalignment="right",
                fontsize=label_fontsize,
            )

    if xlabel is None:
        unit = set(doc.get("magres_units", {}).get("ms", "ppm") for doc in magres)
        if len(unit) > 1:
            raise RuntimeError(
                f"Multiple incompatible units found for chemical shift: {unit}"
            )
        unit = list(unit)[0]
        if magres_key == "chemical_shielding_iso":
            xlabel = f"{species_label}: Isotropic chemical shielding $\\sigma_\\mathrm{{iso}}$ ({unit})"
        elif magres_key == "chemical_shift_iso":
            xlabel = f"{species_label}: Isotropic chemical shift $\\delta_\\mathrm{{iso}}$ ({unit})"
        elif magres_key == "chemical_shift_aniso":
            xlabel = f"{species_label}: Anisotropic chemical shift $\\Delta $ ({unit})"
        elif magres_key == "chemical_shift_asymmetry":
            xlabel = f"{species_label}: Chemial shift asymmetry, $\\eta$"

    ax.set_xlabel(xlabel)
    ax.set_ylabel("Intensity (arb. units)")

    if len(magres) > 1:
        ax.set_yticks([])
    else:
        ax.set_yticks(np.linspace(0, 1, 5, endpoint=True))

    ax.set_ylim(-0.1, 1.1 * len(magres))
    if invert_xaxis:
        ax.invert_xaxis()

    if savefig:
        plt.savefig(savefig)
        print("Wrote {}".format(savefig))

    elif show:
        plt.show()

    return ax



# coding: utf-8
# Distributed under the terms of the MIT License.

""" This module contains various utility functions that
are used liberally throughout matador.

"""


__all__ = []
__author__ = "Matthew Evans"
__maintainer__ = "Matthew Evans"



# coding: utf-8
# Distributed under the terms of the MIT License.

""" This module contains functionality for wrapping matador
data dictionaries as :obj:`Crystal` objects, which possess
several convenience methods.

"""


__all__ = ["Crystal"]

__author__ = "Matthew Evans"
__maintainer__ = "Matthew Evans"


from .crystal import Crystal



# coding: utf-8
# Distributed under the terms of the MIT License.

""" This file implements plotting routines specifically
for the PDF and PDFOverlap objects defined in the
matador.fingerprints.pdf module.

"""

import numpy as np

from matador.crystal import Crystal
from matador.plotting.plotting import plotting_function
from matador.utils.cell_utils import get_space_group_label_latex
from matador.utils.chem_utils import get_formula_from_stoich

__all__ = [
    "plot_pdf",
    "plot_projected_pdf",
    "plot_diff_overlap",
    "plot_projected_diff_overlap",
]


@plotting_function
def plot_pdf(
    pdfs,
    labels=None,
    r_min=None,
    r_max=None,
    offset=1.2,
    text_offset=(0.0, 0.0),
    legend=False,
    annotate=True,
    figsize=None,
    ax=None,
    projected=False,
    colour_labels=True,
    **kwargs,
):
    """Plot PDFs.

    Parameters:
        pdfs (list of matador.fingerprints.pdf.PDF or matador.crystal.Crystal or dict):
            the PDFs to plot, as a list of PDF or Crystal objects, or a matador document.

    Keyword arguments:
        labels (list of str): labels to add to the PDF plot.
        offset (float): amount by which to separate the PDFs in the plot. A value of 1
            will separate by the maximum intensity across the PDFs. Default is 1.5.
        text_offset (tuple of float): two float values to move annotations around relative
            to the base of their corresponding PDF, in units of (Angstrom, max_gr).
        r_max (float): the radius to plot out to. Default is the minmax(radius across
            all PDFs).
        annotate (bool): whether or not to apply the PDF labels as an annotation.
        legend (bool): whether or not to apply the PDF labels as a legend.
        figsize (tuple of float): matplotlib figure size. Default scales with number of PDFs.
        ax (matplotlib.Axis): optional axis object to plot on.
        projected (list(str)): if provided or True, will plot the PDFs projected onto the given keys.
        colour_labels (bool): whether to colour the labels based on the PDF's colour.

    Returns:
        matplotlib.pyplot.Axes: axis object which can be modified further.

    """

    import matplotlib.pyplot as plt
    from matador.utils.viz_utils import get_element_colours

    if not isinstance(pdfs, list):
        pdfs = [pdfs]
    if labels is not None and not isinstance(labels, list):
        labels = [labels]

    if figsize is None and ax is None:
        _user_default_figsize = plt.rcParams.get("figure.figsize", (8, 6))
        height = len(pdfs) * max(0.5, _user_default_figsize[1] / 1.5 / len(pdfs))
        figsize = (_user_default_figsize[0], height)

    if ax is None:
        fig, ax = plt.subplots(figsize=figsize)

    if labels is not None and len(labels) != len(pdfs):
        raise RuntimeError("Wrong number of labels {} for PDFs.".format(labels))

    if isinstance(pdfs[0], Crystal):
        gr_max = max(np.max(pdf.pdf.gr) for pdf in pdfs)
        _r_max = min(np.max(pdf.pdf.r_space) for pdf in pdfs)
    elif isinstance(pdfs[0], dict):
        gr_max = max(np.max(pdf["pdf"].gr) for pdf in pdfs)
        _r_max = min(np.max(pdf["pdf"].r_space) for pdf in pdfs)
    else:
        gr_max = max(np.max(pdf.gr) for pdf in pdfs)
        _r_max = min(np.max(pdf.r_space) for pdf in pdfs)
    abs_offset = offset * gr_max

    if r_max is None:
        r_max = _r_max
    if r_min is None:
        r_min = 0.0

    ax.set_ylabel("Pair distribution function, $g(r)$")
    ax.get_yaxis().set_ticks([])
    ax.set_xlim(r_min, r_max + 0.5)

    projected_keys = set()
    labelled_keys = set()

    for ind, pdf in enumerate(pdfs):

        if isinstance(pdf, Crystal):
            pdf = pdf.pdf
        elif isinstance(pdf, dict) and "pdf" in pdf:
            pdf = pdf["pdf"]

        if labels:
            label = labels[ind]
        else:
            label = f"{get_formula_from_stoich(pdf.stoichiometry, tex=True)}-{get_space_group_label_latex(pdf.spg)}"

        if projected:
            if isinstance(projected, bool):
                keys = [key for key in pdf.elem_gr]
                [projected_keys.add(key) for key in keys]
            else:
                for key in projected:
                    if key in pdf.elem_gr:
                        projected_keys.add(key)
                    elif ([key[1], key[0]]) in pdf.elem_gr:
                        projected_keys.add((key[1], key[0]))

            for key in projected_keys:
                if key in pdf.elem_gr:
                    _label = None
                    if key not in labelled_keys:
                        if len(key) == 2:
                            _label = f"{key[0]}-{key[1]}"
                        else:
                            _label = f"{key[0]}-{key[0]}"
                        labelled_keys.add(key)
                    if len(key) == 2:
                        color = (
                            np.array(get_element_colours()[key[0]])
                            + np.array(get_element_colours()[key[1]])
                        ) / 2
                    else:
                        color = get_element_colours()[key[0]]
                    ax.plot(
                        pdf.r_space,
                        pdf.elem_gr[key] + abs_offset * ind,
                        label=_label,
                        c=color,
                    )

        else:
            color = next(ax._get_lines.prop_cycler)["color"]
            ax.plot(pdf.r_space, pdf.gr + abs_offset * ind, label=label, color=color)
        if text_offset is not None:
            text_x = text_offset[0] + r_min
        if text_offset is not None:
            text_y = abs_offset * ind + text_offset[1] * gr_max
        if label is not None and annotate:
            text_color = None
            if colour_labels:
                text_color = color
            ax.text(text_x, text_y, label, color=text_color)

    ax.set_ylim(-gr_max * 0.2, offset * gr_max * len(pdfs))

    ax.set_xlabel("$r$ ($\\AA$)")

    if legend or projected:
        legend = ax.legend(c="upper right")

    return ax


@plotting_function
def plot_projected_pdf(*args, **kwargs):
    """DEPRECATED"""
    return plot_pdf(*args, **kwargs)


@plotting_function
def plot_diff_overlap(pdf_overlap):
    """Simple plot for comparing two PDFs.

    Parameters:
        pdf_overlap (matador.fingerprints.pdf.PDFOverlap): the
        overlap object to plot.

    """
    import matplotlib.pyplot as plt
    import matplotlib.gridspec as gridspec
    import numpy as np

    plt.figure()
    gs = gridspec.GridSpec(2, 1, height_ratios=[2, 1])
    gs.update(hspace=0)

    ax1 = plt.subplot(gs[0])
    ax2 = plt.subplot(gs[1], sharex=ax1)

    ax2.set_xlabel("$r$ (\\AA)")
    ax1.set_ylabel("$g(r)$")
    ax2.set_ylabel("$g_a(r) - g_b(r)$")
    ax2.axhline(0, ls="--", c="k", lw=0.5)
    ax1.set_xlim(0, np.max(pdf_overlap.fine_space))

    ax1.plot(
        pdf_overlap.fine_space, pdf_overlap.fine_gr_a, label=pdf_overlap.pdf_a.label
    )
    ax1.plot(
        pdf_overlap.fine_space, pdf_overlap.fine_gr_b, label=pdf_overlap.pdf_b.label
    )

    plt.setp(ax1.get_xticklabels(), visible=False)
    ax2.set_ylim(-0.5 * ax1.get_ylim()[1], 0.5 * ax1.get_ylim()[1])

    ax1.legend(loc=0)
    ax2.plot(pdf_overlap.fine_space, pdf_overlap.overlap_fn, ls="-")
    ax2.set_ylim(ax1.get_ylim()[1], ax1.get_ylim()[1])


@plotting_function
def plot_projected_diff_overlap(pdf_overlap):
    """Simple plot for comparing two PDFs.

    Parameters:
        pdf_overlap (matador.fingerprints.pdf.PDFOverlap): the
            overlap object to plot.

    """
    import matplotlib.pyplot as plt
    import numpy as np
    import matplotlib.gridspec as gridspec

    plt.figure()
    gs = gridspec.GridSpec(2, 1, height_ratios=[2, 1])
    gs.update(hspace=0)

    ax1 = plt.subplot(gs[0])
    ax2 = plt.subplot(gs[1], sharex=ax1)
    ax2.set_xlabel("$r$ (\\AA)")
    ax1.set_ylabel("$g(r)$")
    ax2.set_ylabel("$g_a(r) - g_b(r)$")
    ax2.axhline(0, ls="--", c="k", lw=0.5)
    ax1.set_xlim(0, np.max(pdf_overlap.fine_space))
    for _, key in enumerate(pdf_overlap.fine_elem_gr_a):
        ax1.plot(
            pdf_overlap.fine_space,
            pdf_overlap.fine_elem_gr_a[key],
            label="-".join(key) + " {}".format(pdf_overlap.pdf_a.label),
        )
        ax1.plot(
            pdf_overlap.fine_space,
            pdf_overlap.fine_elem_gr_b[key],
            label="-".join(key) + " {}".format(pdf_overlap.pdf_b.label),
            ls="--",
        )
        ax2.plot(
            pdf_overlap.fine_space,
            pdf_overlap.fine_elem_gr_a[key] - pdf_overlap.fine_elem_gr_b[key],
            label="-".join(key) + " diff",
        )
    plt.setp(ax1.get_xticklabels(), visible=False)
    ax2.set_ylim(ax1.get_ylim()[1], ax1.get_ylim()[1])
    ax1.legend(loc=0)
    ax2.legend(loc=2)



from collections import defaultdict

import matplotlib.colors
import matplotlib.pyplot as plt
import numpy as np

from matador.plotting.plotting import get_linear_cmap, plotting_function


@plotting_function
def plot_relative_diffs(
    structure_map,
    structure_comparator,
    categories,
    field,
    ax=None,
    x_field="concentration",
    category_labels=None,
    field_label=None,
    x_field_label=None,
    colour_by_index=1,
    cbar_label=None,
    diff_type="rel",
    invert=False,
    colour_by=None,
    figsize=None,
    colourbar=True,
):
    """Plot relative and absolute differences between fields for a set of structures
    computed at different levels of theory/accuracy.

    """
    if ax is None:
        _, axes = plt.subplots(
            1, len(categories) - 1, figsize=figsize, squeeze=True, sharey=True
        )
    else:
        axes = [ax]
    if category_labels is None:
        category_labels = categories
    if field_label is None:
        field_label = f"change in {field.replace('_', ' ')}"
        if diff_type == "rel":
            field_label = "% " + field_label
    if x_field_label is None:
        x_field_label = x_field.replace("_", " ")

    if len(categories) > 2:
        raise RuntimeError("Cannot plot more than two categories simultaneously.")

    try:
        len(axes)
    except Exception:
        axes = [axes]

    for ax in axes:
        ax.set_ylabel(field_label)
        ax.set_xlabel(x_field_label)
        ax.axhline(0, ls="--", c="grey", alpha=0.5)
    x = []
    for s in structure_comparator:
        x.append(structure_map[s][categories[0]][x_field])
    if colour_by is not None:
        colour_by_x = []
        for s in structure_comparator:
            colour_by_x.append(
                structure_comparator[s]
                .get(categories[colour_by_index], {})
                .get(colour_by)
            )
    else:
        colour_by_x = None
    ys = defaultdict(list)
    if diff_type in ("abs", "rel"):
        for s in structure_comparator:
            for label in categories[1:]:
                ys[label].append(
                    structure_comparator[s].get(label, {}).get(f"{diff_type}_{field}")
                )
    else:
        for s in structure_comparator:
            for label in categories[1:]:
                ys[label].append(structure_comparator[s].get(label, {}).get(field))

    for ind, (label, y) in enumerate(ys.items()):

        if diff_type == "rel":
            y = [100 * yv if yv is not None else None for yv in y]
        if invert:
            y = [-1 * yv if yv is not None else None for yv in y]

        colour_by_x = np.array(colour_by_x, dtype=np.float64)
        x = np.array(x)[np.where(~np.isnan(colour_by_x))]
        y = np.array(y)[np.where(~np.isnan(colour_by_x))]
        colour_by_x = colour_by_x[np.where(~np.isnan(colour_by_x))]

        order = np.argsort(colour_by_x)[::-1]
        y = y[order]
        x = x[order]
        colour_by_x = colour_by_x[order]

        norm = None
        ticks = None
        default_cmap = "viridis"
        extend = None
        if colour_by == "hull_distance":
            norm = matplotlib.colors.LogNorm(0.01, 1, clip=True)
            ticks = [0.0, 0.01, 0.02, 0.04, 0.08, 0.16, 0.32, 0.64]
            colours = list(plt.rcParams["axes.prop_cycle"].by_key()["color"])
            default_cmap = get_linear_cmap(colours[1:4], list_only=False)
            extend = "both"
            hull_points = np.where(colour_by_x <= 0.0)

            if x_field == "concentration":
                x = np.array(x).flatten()
                y = np.array(y).flatten()

                axes[ind].plot(
                    x[hull_points][np.argsort(x[hull_points])],
                    y[hull_points][np.argsort(x[hull_points])],
                    c="k",
                    zorder=1e10,
                    lw=2,
                    markeredgecolor="k",
                    markerfacecolor=colours[1],
                    marker="o",
                    markersize=np.sqrt(40),
                    markeredgewidth=1.5,
                    alpha=1,
                )

            else:
                axes[ind].scatter(
                    np.array(x)[hull_points],
                    np.array(y)[hull_points],
                    c=colours[1],
                    zorder=1e10,
                    lw=1.5,
                    edgecolor="k",
                    norm=norm,
                    alpha=1,
                )

        scatter = axes[ind].scatter(
            x,
            y,
            cmap=default_cmap,
            c=colour_by_x,
            zorder=-1,
            alpha=0.8,
        )
        axes[ind].set_title(f"{category_labels[0]} vs {category_labels[ind+1]}")

    if colour_by_x is not None and colourbar:
        cbar = plt.colorbar(scatter, aspect=30, pad=0.02, extend=extend, ticks=ticks)
        if ticks:
            cbar.ax.set_yticklabels(ticks)
            cbar.ax.tick_params(which="minor", length=0)
        if cbar_label is None and colour_by is not None:
            cbar_label = colour_by.replace("_", " ")
        cbar.set_label(cbar_label)

    if len(axes) == 1:
        axes = axes[0]

    return axes



# coding: utf-8
# Distributed under the terms of the MIT License.

""" This file implements the "matador" command line. """


from matador.query import DBQuery
from matador.hull import QueryConvexHull
from matador.utils.print_utils import print_failure, print_warning, print_notify
from matador.db import make_connection_to_collection
from matador.config import load_custom_settings


class MatadorCommandLine:
    """Class that implements the command-line interface to a MongoDB
    structure repository.

    """

    def __init__(self, *args, **kwargs):
        """Initialise the query with command line arguments and return
        results.

        """
        # read args
        self.kwargs = kwargs
        self.args = vars(args[0])
        self.args["no_quickstart"] = self.kwargs.get("no_quickstart")
        self.argstr = kwargs.get("argstr")

        file_exts = ["cell", "res", "pdb", "markdown", "latex", "param", "xsf"]
        self.export = any([self.args.get(ext) for ext in file_exts])

        self.subcommand = self.args.pop("subcmd")

        if self.subcommand != "import":
            self.settings = load_custom_settings(
                config_fname=self.args.get("config"),
                debug=self.args.get("debug"),
                no_quickstart=self.args.get("no_quickstart"),
            )
            result = make_connection_to_collection(
                self.args.get("db"),
                check_collection=(self.subcommand != "stats"),
                mongo_settings=self.settings,
            )
            self.client, self.db, self.collections = result

        if self.subcommand == "stats":
            self.stats()

        try:
            if self.subcommand == "import":
                from matador.db import Spatula

                self.importer = Spatula(self.args)

            if self.subcommand == "query":
                self.query = DBQuery(self.client, self.collections, **self.args)
                self.cursor = self.query.cursor

            if self.subcommand == "swaps":
                from matador.swaps import AtomicSwapper

                self.query = DBQuery(self.client, self.collections, **self.args)
                if self.args.get("hull_cutoff") is not None:
                    self.hull = QueryConvexHull(query=self.query, **self.args)
                    self.swapper = AtomicSwapper(self.hull.hull_cursor, **self.args)
                else:
                    self.swapper = AtomicSwapper(self.query.cursor, **self.args)
                self.cursor = self.swapper.cursor

            if self.subcommand == "refine":
                from matador.db import Refiner

                self.query = DBQuery(self.client, self.collections, **self.args)
                if self.args.get("hull_cutoff") is not None:
                    self.hull = QueryConvexHull(self.query, **self.args)
                    self.refiner = Refiner(
                        self.hull.cursor, self.query.repo, **self.args
                    )
                else:
                    self.refiner = Refiner(
                        self.query.cursor, self.query.repo, **self.args
                    )

                self.cursor = self.refiner.cursor

            if self.subcommand == "hull" or self.subcommand == "voltage":
                self.hull = QueryConvexHull(
                    **self.args,
                    voltage=self.subcommand == "voltage",
                    client=self.client,
                    collections=self.collections
                )
                self.cursor = self.hull.hull_cursor

            if self.subcommand == "changes":
                from matador.db import DatabaseChanges

                if len(self.collections) != 1:
                    raise SystemExit(
                        "Cannot view changes of more than one collection at once."
                    )
                if self.args.get("undo"):
                    action = "undo"
                else:
                    action = "view"
                changeset = self.args.get("changeset")
                if changeset is None:
                    changeset = 0
                DatabaseChanges(
                    [key for key in self.collections][0],
                    changeset_ind=changeset,
                    action=action,
                    mongo_settings=self.settings,
                    override=kwargs.get("no_quickstart"),
                )

            if self.subcommand == "hulldiff":
                from matador.hull.hull_diff import diff_hulls

                if self.args.get("compare") is None:
                    raise SystemExit(
                        "Please specify which hulls to query with --compare."
                    )
                diff_hulls(self.client, self.collections, **self.args)

            if self.export and self.cursor:
                from matador.export import query2files

                if self.args.get("write_n") is not None:
                    self.cursor = [
                        doc
                        for doc in self.cursor
                        if len(doc["stoichiometry"]) == self.args.get("write_n")
                    ]
                if not self.cursor:
                    print_failure("No structures left to export.")
                query2files(
                    self.cursor,
                    **self.args,
                    argstr=self.argstr,
                    subcmd=self.subcommand,
                    hash_dupe=True
                )

            if self.args.get("view"):
                from matador.utils.viz_utils import viz

                if self.args.get("top") is None:
                    self.top = len(self.cursor)
                else:
                    self.top = self.args.get("top")
                if len(self.cursor[: self.top]) > 10:
                    from time import sleep

                    print_warning(
                        "WARNING: opening {} files with ase-gui...".format(
                            len(self.cursor)
                        )
                    )
                    print_warning("Please kill script within 3 seconds if undesired...")
                    sleep(3)
                if len(self.cursor[: self.top]) > 20:
                    print_failure(
                        "You will literally be opening that many windows, "
                        + "I'll give you another 5 seconds to reconsider..."
                    )

                    sleep(5)
                    print_notify("It's your funeral...")
                    sleep(1)
                for doc in self.cursor[: self.top]:
                    viz(doc)

            if self.subcommand != "import":
                self.client.close()

        except (RuntimeError, SystemExit, KeyboardInterrupt) as oops:
            if isinstance(oops, RuntimeError):
                print_failure(oops)
            elif isinstance(oops, SystemExit):
                print_warning(oops)
            try:
                self.client.close()
            except AttributeError:
                pass
            raise oops

    def print_report(self):
        """Print spatula report on current database."""
        try:
            report = self.report.find_one()
            print(
                "Database last modified on",
                report["last_modified"],
                "with matador",
                report["version"] + ".",
            )
        except Exception:
            print_warning(
                "Failed to print database report: spatula is probably running!"
            )

    def stats(self):
        """Print some useful stats about the database."""
        if self.args.get("list"):
            print_notify(
                str(len(self.db.list_collection_names()))
                + " collections found in database:\n"
            )
            collstats_list = []
            for name in self.db.list_collection_names():
                collstats_list.append(self.db.command("collstats", name))
                collstats_list[-1]["name"] = name
            collstats_list = sorted(
                collstats_list, key=lambda k: k["count"], reverse=True
            )
            print("\t{:^20}\t{:^20}".format("Name", "Number of structures"))
            for collection in collstats_list:
                if not collection["name"].startswith("__"):
                    print(
                        "\t{:<20}\t{:>20d}".format(
                            collection["name"], collection["count"]
                        )
                    )
            print("\n")
        elif self.args.get("delete"):
            target = self.args.get("db")
            if isinstance(target, list) and len(target) == 1:
                target = target[0]
            else:
                raise SystemExit("I will only delete one collection at a time...")
            if target is None:
                raise SystemExit("Please specify a collection to delete.")
            if target not in self.db.list_collection_names():
                raise SystemExit("No collection named {} was found".format(target))

            from getpass import getuser

            user = getuser()
            if user not in target:
                raise SystemExit(
                    "I cannot delete a collection that's name does not start with "
                    "your username, {}".format(user)
                )
            stats = self.db.command("collstats", target)

            if self.args.get("no_quickstart"):
                answer = "y"
            else:
                answer = input(
                    "Are you sure you want to delete collection {} containing {} "
                    "structures? [y/n]\n".format(target, stats["count"])
                )
            if answer.lower() == "y":
                if target == "repo":
                    raise SystemExit("I'm sorry Dave, I'm afraid I can't do that...")
                print("Deleting collection {}...".format(target))
                self.db[target].drop()
                print("and its changelog...")
                self.db["__changelog_{}".format(target)].drop()
            else:
                raise SystemExit("Nevermind then!")

        else:
            comp_list = dict()
            stats_dict = dict()
            stats_dict["count"] = 0
            stats_dict["avgObjSize"] = 0
            stats_dict["storageSize"] = 0
            stats_dict["totalIndexSize"] = 0
            for collection in self.collections:
                db_stats_dict = self.db.command("collstats", collection)
                stats_dict["count"] += db_stats_dict["count"]
                stats_dict["avgObjSize"] += db_stats_dict["avgObjSize"]
                stats_dict["storageSize"] += db_stats_dict["storageSize"]
                stats_dict["totalIndexSize"] += db_stats_dict["totalIndexSize"]
            print(
                (
                    "The collection(s) queried in {} contain {} structures at {:.1f} kB each "
                    "totalling {:.1f} MB with a further {:.1f} MB of indexes."
                ).format(
                    self.db.name,
                    stats_dict["count"],
                    stats_dict["avgObjSize"] / (1024),
                    stats_dict["storageSize"] / (1024**2),
                    stats_dict["totalIndexSize"] / (1024**2),
                )
            )
            for collname in self.collections:
                cursor = self.collections[collname].find()
                for doc in cursor:
                    temp = ""
                    for ind, elem in enumerate(sorted(doc["stoichiometry"])):
                        temp += str(elem[0])
                        if ind != len(doc["stoichiometry"]) - 1:
                            temp += "+"
                    if temp not in comp_list:
                        comp_list[temp] = 0
                    comp_list[temp] += 1
            keys = list(comp_list.keys())
            vals = list(comp_list.values())
            comp_list = list(zip(keys, vals))
            comp_list.sort(key=lambda t: t[1], reverse=True)
            small_count = 0
            first_ind = 1000
            cutoff = 100
            for ind, comp in enumerate(comp_list):
                if comp[1] < cutoff:
                    if ind < first_ind:
                        first_ind = ind
                    small_count += comp[1]
            comp_list = comp_list[:first_ind]
            comp_list.append(["others < " + str(cutoff), small_count])
            comp_list.sort(key=lambda t: t[1], reverse=True)
            try:
                from ascii_graph import Pyasciigraph
                from ascii_graph.colors import Gre, Blu, Red
                from ascii_graph.colordata import hcolor
            except ImportError:
                print("ascii_graph dependency not found, not creating histogram.")
            else:
                graph = Pyasciigraph(line_length=80, multivalue=False)
                thresholds = {
                    int(stats_dict["count"] / 40): Gre,
                    int(stats_dict["count"] / 10): Blu,
                    int(stats_dict["count"] / 4): Red,
                }
                data = hcolor(comp_list, thresholds)
                for line in graph.graph(label=None, data=data):
                    print(line)
                print("\n")

                for comp in comp_list:
                    print(comp)


def main(no_quickstart=False):
    """Parse all user args and construct a MatadorCommandLine object.

    Keyword arguments:
        no_quickstart: no_quickstart all stdin with sensible defaults.

    """
    import argparse
    from sys import argv
    from matador import __version__, script_epilog

    parser = argparse.ArgumentParser(
        prog="matador",
        description="MATerial and Atomistic Database Of Refined structures.",
        epilog=script_epilog,
    )
    parser.add_argument(
        "--version", action="version", version="matador version " + __version__ + "."
    )

    # define subparsers for self.subcommands
    subparsers = parser.add_subparsers(
        title="self.subcommands", description="valid sub-commands", dest="subcmd"
    )

    # define parent parser for global arguments
    global_flags = argparse.ArgumentParser(add_help=False)

    # common arguments to all self.subcommands
    global_flags.add_argument(
        "--db", nargs="+", help="choose which collection to query"
    )
    global_flags.add_argument(
        "--debug", action="store_true", help="enable debug printing throughout code."
    )
    global_flags.add_argument(
        "-conf",
        "--config",
        type=str,
        help="specify custom location of matador config file."
        "(DEFAULT: $MATADOR_ROOT/config/matador_conf.json)",
    )
    global_flags.add_argument("--devel", action="store_true", help="test devel code.")
    global_flags.add_argument(
        "--profile", action="store_true", help="run code profiler."
    )
    global_flags.add_argument(
        "-q", "--quiet", action="store_true", help="redirect most output to /dev/null."
    )

    # define all other flags by group
    structure_flags = argparse.ArgumentParser(add_help=False)
    structure_flags.add_argument(
        "-c",
        "--composition",
        type=str,
        nargs="+",
        help="find all structures containing exclusively the given "
        "elements, e.g. LiSi. Macros defined for groups [I]-[VII] "
        "[Tran] [Lan] and [Act], used with square brackets.",
    )
    structure_flags.add_argument(
        "-int",
        "--intersection",
        action="store_true",
        help="query the intersection of compositions instead of the union "
        "e.g. -c LiSnS -int queries Li, Sn, S, LiSn, LiS & LiSnS.",
    )
    structure_flags.add_argument(
        "-n",
        "--num_species",
        type=int,
        help="find all structures containing a certain number of species.",
    )
    structure_flags.add_argument(
        "-f",
        "--formula",
        type=str,
        nargs="+",
        help="query a particular chemical formula, e.g. GeTeSi3",
    )
    structure_flags.add_argument(
        "-i",
        "--id",
        type=str,
        nargs="+",
        help="specify a particular structure by its text_id",
    )
    structure_flags.add_argument(
        "-ac",
        "--calc_match",
        action="store_true",
        help="display calculations of the same accuracy as specified id",
    )
    structure_flags.add_argument(
        "-kpttol",
        "--kpoint_tolerance",
        type=float,
        help="kpoint tolerance for calculation matches (DEFAULT: +/- 0.01 1/Å)",
    )
    structure_flags.add_argument(
        "-presstol",
        "--pressure_tolerance",
        type=float,
        help="pressure tolerance for calculation matches (DEFAULT: +/- 0.5 GPa)",
    )
    structure_flags.add_argument(
        "-z",
        "--num_fu",
        type=int,
        help="query a calculations with more than n formula units",
    )
    structure_flags.add_argument(
        "-sg", "--space_group", help="query a particular space group"
    )
    structure_flags.add_argument(
        "-u",
        "--uniq",
        type=float,
        nargs="?",
        const=0.1,
        help="float, return only unique structures (filtered by PDF "
        "overlap), to this tolerance (DEFAULT: 0.1)",
    )
    structure_flags.add_argument(
        "-p",
        "--pressure",
        type=float,
        help="specify an isotropic external pressure to search for, e.g. 10 (GPa)",
    )
    structure_flags.add_argument(
        "-pf",
        "--partial-formula",
        action="store_true",
        help="stoichiometry/composition queries will include other unspecified species, e.g. "
        "-pf search for Li will query any structure containing Li, not just pure Li.",
    )
    structure_flags.add_argument(
        "--tags", nargs="+", type=str, help=("search for manual tags")
    )
    structure_flags.add_argument(
        "--doi", type=str, help=("search for DOI in format xxxx/xxxx")
    )
    structure_flags.add_argument(
        "-icsd",
        "--icsd",
        type=int,
        const=0,
        nargs="?",
        help=("search for an ICSD CollCode"),
    )
    structure_flags.add_argument(
        "-ss",
        "--src_str",
        type=str,
        help=("search for a string inside the structure sources"),
    )
    structure_flags.add_argument(
        "-root",
        "--root_src",
        type=str,
        help=("search for a root_source string of the structure"),
    )
    structure_flags.add_argument(
        "-encap",
        "--encapsulated",
        action="store_true",
        help="query only structures encapsulated in a carbon nanotube.",
    )
    structure_flags.add_argument(
        "-cntr",
        "--cnt_radius",
        type=float,
        help="specify the radius of the encapsulating nanotube to within 0.01 Å",
    )
    structure_flags.add_argument(
        "-cntv",
        "--cnt_vector",
        type=int,
        nargs="+",
        help="specify the chiral vector of the encapsulating nanotube",
    )
    structure_flags.add_argument(
        "-ecut",
        "--cutoff",
        type=float,
        nargs="+",
        help="specify the min. and optionally max. planewave cutoff.",
    )
    structure_flags.add_argument(
        "-geom",
        "--geom_force_tol",
        type=float,
        nargs="+",
        help="force tolerance in eV/Å to query for calc matches.",
    )
    structure_flags.add_argument(
        "-grid",
        "--grid_scale",
        type=float,
        nargs="+",
        help="grid scale to query for calc matches.",
    )
    structure_flags.add_argument(
        "-finegrid",
        "--fine_grid_scale",
        type=float,
        nargs="+",
        help="fine grid scale to query for calc matches.",
    )
    structure_flags.add_argument(
        "--sedc",
        type=str,
        help="specify the dispersion correction scheme, e.g. TS or null.",
    )
    structure_flags.add_argument(
        "-xc",
        "--xc_functional",
        type=str,
        help="specify an xc-functional to query (case-insensitive).",
    )
    structure_flags.add_argument(
        "-kpts",
        "--mp_spacing",
        type=float,
        help="specify an MP grid spacing in 2π/Å units, e.g. 0.05, will return all values "
        "structures with value within --kpt_tol",
    )
    structure_flags.add_argument(
        "--spin",
        type=str,
        help="specifiy whether to query non-spin-polarized (0) calcs or spin polarized calcs "
        "(!=1), or lump them both together with `any`",
    )
    structure_flags.add_argument(
        "--loose",
        action="store_true",
        help="loosely matches with calc_match, i.e. only matches pspot and xc_functional",
    )
    structure_flags.add_argument(
        "--ignore_warnings",
        action="store_true",
        help="includes possibly bad structures",
    )
    structure_flags.add_argument(
        "--field", type=str, action="append", help="name of arbitrary field to query"
    )
    structure_flags.add_argument(
        "--filter",
        nargs="+",
        action="append",
        help="specify either float [min, max] or a string/float value.",
    )

    material_flags = argparse.ArgumentParser(add_help=False)
    material_flags.add_argument(
        "-hc",
        "--hull_cutoff",
        type=float,
        help="return only structures within a certain distance from hull in eV/atom",
    )
    material_flags.add_argument(
        "-lc",
        "--label_cutoff",
        nargs="+",
        type=float,
        help="label only structures within a certain distance from hull in eV/atom",
    )
    material_flags.add_argument(
        "--biggest",
        action="store_true",
        help="use the largest subset of structures to create a hull",
    )
    material_flags.add_argument(
        "--volume",
        action="store_true",
        help="plot a volume curve from convex hull (currently limited to binaries)",
    )
    material_flags.add_argument(
        "--chempots",
        type=float,
        nargs="+",
        help="manually specify chem pots as enthalpy per atom for a rough hull.",
    )

    plot_flags = argparse.ArgumentParser(add_help=False)
    plot_flags.add_argument(
        "--pdf", action="store_true", help="save pdf rather than showing plot in X"
    )
    plot_flags.add_argument(
        "--png", action="store_true", help="save png rather than showing plot in X"
    )
    plot_flags.add_argument(
        "--csv", action="store_true", help="save plotting data to separate csv files"
    )
    plot_flags.add_argument("--labels", action="store_true", help="label hull plots")
    plot_flags.add_argument(
        "--svg", action="store_true", help="save svg rather than showing plot in X"
    )
    plot_flags.add_argument(
        "--subplot", action="store_true", help="plot combined hull and voltage graph"
    )
    plot_flags.add_argument("--no_plot", action="store_true", help="suppress plotting")
    plot_flags.add_argument(
        "--capmap", action="store_true", help="plot heat map of gravimetric capacity"
    )
    plot_flags.add_argument(
        "--sampmap", action="store_true", help="plot heat map of concentration sampling"
    )
    plot_flags.add_argument(
        "--efmap", action="store_true", help="plot heat map of formation energy"
    )
    plot_flags.add_argument(
        "--pathways",
        action="store_true",
        help="plot line from stable B_x C_y to pure A in ABC ternary.",
    )
    plot_flags.add_argument(
        "--expt",
        type=str,
        help="enter experimental voltage curve .csv file for plotting.",
    )
    plot_flags.add_argument(
        "--expt_label", type=str, help="label for experimental data on voltage curve."
    )

    import_flags = argparse.ArgumentParser(add_help=False)
    import_flags.add_argument(
        "-d",
        "--dryrun",
        action="store_true",
        help="run the importer without connecting to the database",
    )
    import_flags.add_argument(
        "-v", "--verbosity", type=int, help="enable verbose output", default=0
    )
    import_flags.add_argument(
        "-f", "--force", action="store_true", help="override main database protection"
    )
    import_flags.add_argument(
        "-t",
        "--tags",
        nargs="+",
        type=str,
        help="set user tags, e.g. nanotube, project name",
    )
    import_flags.add_argument(
        "--recent_only",
        action="store_true",
        help="sort files by creation date (st_ctime) and "
        "stop importing after a duplicate is found in the database.",
    )
    import_flags.add_argument(
        "-s",
        "--scan",
        action="store_true",
        help="only scan the database for new structures, do not import new structures",
    )
    import_flags.add_argument(
        "-p",
        "--prototype",
        action="store_true",
        help="create a database of prototype structures that contain no DFT calculations",
    )

    changes_flags = argparse.ArgumentParser(add_help=False)
    changes_flags.add_argument(
        "-c", "--changeset", type=int, help="changeset number to query"
    )
    changes_flags.add_argument(
        "-r", "--revert", type=int, help="revert database to specified changeset"
    )
    changes_flags.add_argument(
        "-u", "--undo", action="store_true", help="undo changeset"
    )

    collection_flags = argparse.ArgumentParser(add_help=False)
    collection_flags.add_argument(
        "--to", type=str, help="the text_id of a structure with the desired parameters"
    )
    collection_flags.add_argument(
        "--with",
        type=str,
        help=(
            "the seedname (must be within pwd) of cell and param "
            + "files to use for swaps"
        ),
    )
    collection_flags.add_argument(
        "--prefix",
        type=str,
        help="add a prefix to all file names to write out (auto-appended with an underscore",
    )

    query_flags = argparse.ArgumentParser(add_help=False)

    query_flags.add_argument(
        "-s",
        "--summary",
        action="store_true",
        help="show only the ground state for each stoichiometry.",
    )
    query_flags.add_argument(
        "-t", "--top", type=int, help="number of structures to show/write (DEFAULT: 10)"
    )
    query_flags.add_argument(
        "-dE",
        "--delta_E",
        type=float,
        help="maximum distance from ground state structure to show/write in eV/atom",
    )
    query_flags.add_argument(
        "-d",
        "--details",
        action="store_true",
        help="show as much detail about calculation as possible",
    )
    query_flags.add_argument(
        "-pa",
        "--per_atom",
        action="store_true",
        help="show quantities per atom not per fu.",
    )
    query_flags.add_argument(
        "-ef",
        "--eform",
        action="store_true",
        help="print formation energy not hull distance.",
    )
    query_flags.add_argument(
        "-dt",
        "--time",
        type=int,
        help="query only structures added before this time in days",
    )
    query_flags.add_argument(
        "-avail",
        "--available_values",
        type=str,
        help="list all values of field in query results",
    )
    query_flags.add_argument(
        "--use_source",
        default=False,
        action="store_true",
        help="show the source rather than database ID",
    )
    query_flags.add_argument(
        "--since",
        action="store_true",
        help="query only structures added after time specified by --time in days",
    )
    query_flags.add_argument(
        "--source",
        action="store_true",
        help="print filenames from which structures were wrangled",
    )
    query_flags.add_argument(
        "-v",
        "--view",
        action="store_true",
        help="quickly view a structure/structures with ase-gui",
    )
    query_flags.add_argument(
        "--cell",
        action="store_true",
        help="export query to .cell files in folder name from query string",
    )
    query_flags.add_argument(
        "--param",
        action="store_true",
        help="export query to .param files in folder name from query string",
    )
    query_flags.add_argument(
        "--res",
        action="store_true",
        help="export query to .res files in folder name from query string",
    )
    query_flags.add_argument(
        "--json",
        action="store_true",
        help="export query to raw json files in folder name from query string",
    )
    query_flags.add_argument(
        "--pdb",
        action="store_true",
        help="export query to .pdb files in folder name from query string",
    )
    query_flags.add_argument(
        "--xsf",
        action="store_true",
        help="export query to .xsf files in folder name from query string",
    )
    query_flags.add_argument(
        "--markdown",
        action="store_true",
        help="export query summary to a markdown file",
    )
    query_flags.add_argument(
        "--latex", action="store_true", help="export query summary to a LaTeX table"
    )
    query_flags.add_argument(
        "--write_n", type=int, help="export only those structures with n species"
    )

    swap_flags = argparse.ArgumentParser(add_help=False)
    swap_flags.add_argument(
        "-sw",
        "--swap",
        type=str,
        nargs="+",
        help="swap all atoms in structures from a query from the first n-1 species to the nth, "
        "e.g. -sw NAs will swap all N to As, -sw NAs:LiNa will swap all N to As, and all Li "
        "to Na, and -sw [V]As:[Li,K,Rb]Na will swap all group V elements to As and all of Li,"
        "K and Rb to Na.",
    )
    diff_flags = argparse.ArgumentParser(add_help=False)
    diff_flags.add_argument(
        "-cmp",
        "--compare",
        type=str,
        nargs="+",
        help="diff phase diagrams between two different times, in standard time format, "
        "e.g. `--compare 1y2m5d3h` will compare the present hull with that of 1 year, 2 "
        "months, 5 days and 3 hours ago, and `--compare 3d 2d` will compare three days ago "
        "to two days ago.",
    )

    refine_flags = argparse.ArgumentParser(add_help=False)
    refine_flags.add_argument(
        "-task",
        "--task",
        type=str,
        help=(
            "refine subtask to perform: options are spg, elem_set, tag, doi, source, "
            "pspot or raw or sub"
        ),
    )
    refine_flags.add_argument(
        "-mode",
        "--mode",
        type=str,
        help="mode of refinement: options are display, set and overwrite",
    )
    refine_flags.add_argument(
        "-symprec",
        "--symprec",
        type=float,
        help="spglib symmetry precision for refinement",
    )
    refine_flags.add_argument(
        "--new_tag", type=str, help="new tag to add to structures in query"
    )
    refine_flags.add_argument(
        "--new_doi", type=str, help="new doi to add to structures in query"
    )

    stats_flags = argparse.ArgumentParser(add_help=False)
    stats_flags.add_argument(
        "-l",
        "--list",
        action="store_true",
        help="list all collections, their sizes, and owners",
    )
    stats_flags.add_argument(
        "--delete",
        action="store_true",
        help="try to delete collection specified by --db",
    )

    # define subcommand parsers and their arguments

    # matador stats
    subparsers.add_parser(
        "stats",
        help="print some stats about the database.",
        parents=[global_flags, stats_flags],
    )

    # matador query
    subparsers.add_parser(
        "query",
        help="query and extract structures from the database",
        parents=[global_flags, query_flags, structure_flags],
    )

    # matador import
    subparsers.add_parser(
        "import",
        help="import new structures in folder into database",
        parents=[global_flags, import_flags],
    )

    # matador hull
    subparsers.add_parser(
        "hull",
        help="create a convex hull from query results (currently limited to binaries and ternaries)",
        parents=[
            global_flags,
            structure_flags,
            material_flags,
            plot_flags,
            query_flags,
        ],
    )

    # matador voltage
    subparsers.add_parser(
        "voltage",
        help="plot a voltage curve from query results (currently limited to binaries and ternaries)",
        parents=[
            global_flags,
            structure_flags,
            material_flags,
            plot_flags,
            query_flags,
        ],
    )

    # matador changes
    subparsers.add_parser(
        "changes",
        help="view database changelog or undo additions to database (NB: not deletions!)",
        parents=[global_flags, changes_flags],
    )

    # matador hulldiff
    subparsers.add_parser(
        "hulldiff",
        help="diff two convex hulls with the --compare flag.",
        parents=[
            global_flags,
            structure_flags,
            material_flags,
            plot_flags,
            query_flags,
            diff_flags,
        ],
    )

    # matador swaps
    subparsers.add_parser(
        "swaps",
        help="perform atomic swaps on query results",
        parents=[
            global_flags,
            collection_flags,
            query_flags,
            structure_flags,
            material_flags,
            swap_flags,
        ],
    )

    # matador refine
    subparsers.add_parser(
        "refine",
        help="update structures in the database according to specified --task",
        parents=[
            global_flags,
            query_flags,
            structure_flags,
            refine_flags,
            material_flags,
        ],
    )

    parsed_args = parser.parse_args()
    vars_args = vars(parsed_args)

    # check for inconsistent argument combinations
    if vars_args.get("intersection") and vars_args.get("composition") is None:
        raise SystemExit("--intersection requires --composition.")
    if (
        vars_args.get("subcmd") == "stats"
        and vars_args.get("list")
        and vars_args.get("delete")
    ):
        raise SystemExit("Cannot use -l/--list and --delete")
    if vars_args.get("field") and vars_args.get("filter") is None:
        raise SystemExit("--field requires --filter.")
    if vars_args.get("subcmd") == "hull" and vars_args.get("composition") is None:
        raise SystemExit("hull requires --composition")
    if vars_args.get("calc_match") and vars_args.get("id") is None:
        raise SystemExit(
            "calc_match requires specification of a text_id with -i, exiting..."
        )
    if vars_args.get("profile"):
        import cProfile
        import pstats
        from sys import version_info

        profiler = cProfile.Profile()
        profiler.enable()

    MatadorCommandLine(parsed_args, argstr=argv[1:], no_quickstart=no_quickstart)

    if vars_args.get("profile"):
        profiler.disable()
        fname = "matador-{}-{}.{}.{}".format(
            __version__, version_info.major, version_info.minor, version_info.micro
        )
        profiler.dump_stats(fname + ".prof")
        with open(fname + ".pstats", "w") as fp:
            stats = pstats.Stats(profiler, stream=fp).sort_stats("cumulative")
            stats.print_stats()


if __name__ == "__main__":
    main()



""" This submodule includes some useful utility functions
for exporting data.

"""

import os
import random


def file_writer_function(function):
    """Wrapper for file writers to safely overwrite/hash duplicate files.

    Keyword arguments:
        overwrite (bool): whether or not to overwrite colliding files.
        hash_dupe (bool): whether or not to create a unique filename for
            any colliding files, or just skip writing them.

    """

    from functools import wraps

    @wraps(function)
    def wrapped_writer(doc, path, overwrite=False, hash_dupe=False, **kwargs):
        """Wrap and return the writer function."""
        try:
            overwrite = kwargs.pop("overwrite", overwrite)
            hash_dupe = kwargs.pop("hash_dupe", hash_dupe)
            flines, ext = function(
                doc, path, overwrite=overwrite, hash_dupe=hash_dupe, **kwargs
            )
            if ext is not None and not path.endswith("." + ext):
                path += ".{}".format(ext)

            if os.path.isfile(path):
                if overwrite:
                    os.remove(path)
                elif hash_dupe:
                    path = "{}-{}.{}".format(
                        path.replace(ext, ""), generate_hash(), ext
                    )
                else:
                    return False

            with open(path, "w") as f:
                for line in flines:
                    f.write(line + "\n")

        except Exception as exc:
            raise type(exc)("Failed to write {}: {}".format(path, exc))

    return wrapped_writer


def generate_hash(hash_len=6):
    """Quick hash generator, based on implementation in PyAIRSS by J. Wynn.

    Keyword arguments:
        hash_len (int): desired length of hash.

    """
    hash_chars = [str(x) for x in range(10)] + [chr(97 + i) for i in range(25)]
    _hash = ""
    for _ in range(hash_len):
        _hash += random.choice(hash_chars)
    return _hash


def generate_relevant_path(**args):
    """Generates a suitable path name based on query."""
    dirname = ""
    if args.get("subcmd") is not None:
        dirname += args.get("subcmd") + "-"
    else:
        dirname = "query"
    if args.get("composition") is not None:
        for comp in args["composition"]:
            dirname += comp
    elif args.get("formula") is not None:
        dirname += args.get("formula")[0]
    if args.get("db") is not None:
        dirname += "-" + args.get("db")[0]
    if args.get("swap") is not None:
        for swap in args["swap"]:
            dirname += "-" + swap
        if args.get("hull_cutoff") is not None:
            dirname += "-hull-" + str(args.get("hull_cutoff")) + "eV"
    if args.get("id") is not None:
        dirname += "-" + args.get("id")[0] + "_" + args.get("id")[1]
    dirname = dirname.replace("--", "-")
    return dirname



# coding: utf-8
# Distributed under the terms of the MIT License.

""" This submodule contains a dirty wrapper of ase-gui for quick
visualisation, and nglview wrapper for JupyterNotebook visualisation,
and some colour definitions scraped from VESTA configs.

"""

from typing import Tuple, Optional, Dict, List, Union, Callable, Set, TYPE_CHECKING
from pathlib import Path
from functools import lru_cache

import tqdm
import numpy as np

from matador.utils.ase_utils import doc2ase
from matador.crystal import Crystal

if TYPE_CHECKING:
    import fresnel


def viz(doc):
    """Quick and dirty ase-gui visualisation from matador doc."""
    try:
        from ase.visualize import view
    except ImportError as exc:
        raise ImportError(
            "Optional dependency 'ase' is missing, please install it from PyPI with `pip install ase"
        ) from exc
    view(doc2ase(doc))
    return


@lru_cache(maxsize=1)
def get_element_colours() -> Dict[str, List[float]]:
    """Return RGB element colours from VESTA file.

    The colours file can be specified in the matadorrc.
    If unspecified, the default `../config/vesta_elements.ini` will be used.

    """
    import os
    from matador.config import load_custom_settings, SETTINGS

    default_colours_fname = (
        Path(os.path.realpath(__file__)).parent / "../config/vesta_elements.ini"
    )
    # check if element_colours has been given as an absolute path
    if SETTINGS:
        colours_fname = SETTINGS.get("plotting", {}).get("element_colours")
    else:
        colours_fname = (
            load_custom_settings().get("plotting", {}).get("element_colours")
        )

    if colours_fname is None or not os.path.isfile(colours_fname):
        colours_fname = default_colours_fname

    with open(colours_fname, "r") as f:
        flines = f.readlines()
    element_colours = dict()
    for line in flines:
        line = line.split()
        elem = line[1]
        colour = list(map(float, line[-3:]))
        element_colours[elem] = colour
    return element_colours


@lru_cache(maxsize=1)
def get_element_radii():
    """Read element radii from VESTA config file. The colours file can be
    specified in the matadorrc. If unspecified, the default
    ../config/vesta_elements.ini (relative to this file) will be used.

    """
    import os
    from matador.config import load_custom_settings, SETTINGS

    default_colours_fname = (
        Path(os.path.realpath(__file__)).parent / "../config/vesta_elements.ini"
    )
    # check if element_colours has been given as an absolute path
    if SETTINGS:
        colours_fname = SETTINGS.get("plotting", {}).get("element_colours")
    else:
        colours_fname = (
            load_custom_settings().get("plotting", {}).get("element_colours")
        )

    if colours_fname is None or not os.path.isfile(colours_fname):
        colours_fname = default_colours_fname

    with open(colours_fname, "r") as f:
        flines = f.readlines()
    element_radii = dict()
    for line in flines:
        line = line.split()
        elem = line[1]
        r = float(line[2])
        element_radii[elem] = r

    return element_radii


def nb_viz(doc, repeat=1, bonds=None):
    """Return an ipywidget for nglview visualisation in
    a Jupyter notebook or otherwise.

    Parameters:
        doc (matador.crystal.Crystal / dict): matador document to show.

    Keyword arguments:
        repeat (int): number of periodic images to include.
        bonds (str): custom bond selection.

    """
    try:
        import nglview
    except ImportError as exc:
        raise ImportError(
            "Optional dependency 'nglview' is missing, please install it from PyPI with `pip install nglview"
        ) from exc
    atoms = doc2ase(doc)
    atoms = atoms.repeat((repeat, repeat, repeat))
    view = nglview.show_ase(atoms)
    view.add_unitcell()
    view.remove_ball_and_stick()
    view.add_spacefill(radius_type="vdw", scale=0.3)
    if bonds is None:
        for elements in set(doc["atom_types"]):
            view.add_ball_and_stick(selection="#{}".format(elements.upper()))
    else:
        if not isinstance(bonds, list):
            bonds = [bonds]
        for bond in bonds:
            view.add_ball_and_stick(selection="#{}".format(bond.upper()))
    view.parameters = {"clipDist": 0}
    view.camera = "orthographic"
    view.background = "#FFFFFF"
    view.center()
    return view


def fresnel_view(
    doc: Crystal,
    standardize: Union[bool, float] = True,
    extension: Optional[Tuple[int, int, int]] = None,
    show_bonds: bool = True,
    show_cell: bool = True,
    show_compass: bool = True,
    bond_dict: Optional[Dict[Tuple[str, str], float]] = None,
    images: Union[bool, float] = True,
    pad_cell: bool = True,
    lights: Optional[Callable] = None,
    **camera_kwargs,
) -> "fresnel.Scene":
    """Return a fresnel scene visualising the input crystal.

    The scene can then be rendered/ray-traced with `fresnel.pathtrace` or
    `fresnel.preview`.

    Notes:
        Periodic images of sites will be included in the visualisation if they are
        within the desired distance tolerance of another site in the crystal, and
        are only rendered if they are bonded within the definitions of the `bond_dict`.
        This process requires the construction of two kd-trees, so disabling the `images`
        flag will yield a significant speed-up for large crystal structures.

    Parameters:
        doc: The crystal to visualize.
        standardize: Whether to standardize the crystal before visualising. If a float
            is provided, this will be used for the symmetry tolerance.
        extension: Visualize a supercell with this lattice extension.
        show_bonds: Whether to show bonds in the visualisation.
        show_cell: Whether to show the unit cell in the visualisation.
        show_compass: Whether to show a compass in the visualisation.
        bond_dict: A dictionary of bond lengths to override the default bonds, e.g.,
            `{('C', 'C'): 1.5, ('C', 'H'): 1.0}`.
        images: Whether to show periodic images in the visualisation (if given a float, this will be used
            as the clipping distance).
        pad_cell: Size to aim for all directions to be visualised, e.g. a supercell with minimum side length
            `pad_cell` will be constructed.
        lights: An optional callable that sets the lights for the scene.

    Returns:
        The fresnel scene to render.
    """
    try:
        import fresnel
    except ImportError as exc:
        raise ImportError(
            "Optional dependency 'fresnel' is missing, please install it "
            "from conda forge with `conda install -c conda-forge fresnel`"
        ) from exc

    scene = fresnel.Scene()
    if standardize:
        doc = doc.standardized()

    if show_cell:
        _draw_cell(scene, doc, compass=show_compass)

    if extension is not None:
        if not extension == (1, 1, 1):
            doc = doc.supercell(extension)
    elif pad_cell:
        doc = doc.supercell(target=pad_cell)

    clip = None
    if isinstance(images, float):
        images = True
        clip = images
    elif images and bond_dict:
        clip = max(bond_dict[d] for d in bond_dict)

    species, positions = _get_sites(doc, images=images, clip=clip)

    to_remove = set()
    if show_bonds:
        bond_sets = _draw_bonds(
            scene,
            list(zip(species, positions)),
            bond_dict=bond_dict,
            nsites=len(doc.positions_abs),
            bond_images=False,
        )

        if images:
            image_ind = len(doc.positions_abs)
            for i in range(image_ind, len(positions)):
                if i not in bond_sets:
                    to_remove.add(i)

    positions = [positions[i] for i, p in enumerate(positions) if i not in to_remove]
    species = [species[i] for i, p in enumerate(species) if i not in to_remove]

    _draw_atoms(scene, positions, species)

    a, b, c = doc.lattice_cart
    fit_orthographic_camera(scene, **camera_kwargs)
    if lights is None:
        scene.lights = fresnel.light.lightbox()
    else:
        scene.lights = lights()

    return scene


def _get_sites(
    doc, images: True, clip: float = 3.5
) -> Tuple[List[str], List[Tuple[float, float, float]]]:
    """Return the species and positions of the sites in the crystal, including
    any periodic images within the distance tolerance.

    Parameters:
        doc: The crystal to visualize.
        images: Whether to include periodic images.
        clip: The distance tolerance to other sites for periodic image inclusion.

    Returns:
        A list of species and a list of Cartesian positions.

    """
    positions = []
    species = []
    clip = clip or 3.5
    for ind, pos_cart in enumerate(doc.positions_abs):
        positions.append(pos_cart)
        species.append(doc[ind].species)
    if images:
        image_positions, image_species = get_image_atoms_near_cell_boundaries(
            doc, tolerance=clip
        )
        positions.extend(image_positions)
        species.extend(image_species)

    return species, positions


def _draw_cell(scene, doc, thickness=0.05, compass=True) -> None:
    """Add the unit cell to a fresnel scene.

    Parameters:
        scene: The fresnel scene.
        doc: The crystal to visualize.
        thickness: The line thickness to use.
        compass: Whether to draw the compass of lattice vectors.

    """
    import numpy as np
    import fresnel

    vertices = np.zeros((8, 3), dtype=np.float64)
    vertices[1:4, :] = doc.lattice_cart
    vertices[4, :] = vertices[1, :] + vertices[2, :]
    vertices[5, :] = vertices[1, :] + vertices[3, :]
    vertices[6, :] = vertices[2, :] + vertices[3, :]
    vertices[7, :] = np.sum(vertices[1:4, :], axis=0)
    unit_cell = fresnel.geometry.Cylinder(scene, N=12)
    unit_cell.points[:] = [
        [vertices[0, :], vertices[1, :]],
        [vertices[0, :], vertices[2, :]],
        [vertices[0, :], vertices[3, :]],
        [vertices[1, :], vertices[4, :]],
        [vertices[2, :], vertices[4, :]],
        [vertices[1, :], vertices[5, :]],
        [vertices[3, :], vertices[5, :]],
        [vertices[2, :], vertices[6, :]],
        [vertices[3, :], vertices[6, :]],
        [vertices[4, :], vertices[7, :]],
        [vertices[5, :], vertices[7, :]],
        [vertices[6, :], vertices[7, :]],
    ]
    unit_cell.radius[:] = thickness * np.ones(12)
    unit_cell.material = fresnel.material.Material(
        color=fresnel.color.linear([0, 0, 0])
    )
    unit_cell.material.primitive_color_mix = 1
    unit_cell.material.solid = 1.0

    if compass:
        compass = fresnel.geometry.Cylinder(scene, N=3)
        compass.points[:] = [
            [[0, 0, 0], doc.lattice_cart[ind] / doc.lattice_abc[0][ind]]
            for ind in range(3)
        ]
        compass.color[0] = ([1.0, 0, 0], [1, 0, 0])
        compass.color[1] = ([0, 1.0, 0], [0, 1.0, 0])
        compass.color[2] = ([0, 0, 1.0], [0, 0, 1.0])
        compass.radius[:] = 0.1
        compass.material = fresnel.material.Material(
            color=fresnel.color.linear([0, 0, 0])
        )
        compass.material.primitive_color_mix = 1
        compass.material.solid = 0.0
        compass.material.roughness = 0.5
        compass.material.specular = 0.7
        compass.material.spec_trans = 0.0
        compass.material.metal = 0.0


def _draw_bonds(
    scene: "fresnel.Scene",
    sites: List[Tuple[str, List[List[float]]]],
    bond_dict: Dict[Tuple[str, str], float] = None,
    bond_thickness: float = 0.15,
    nsites: int = None,
    bond_images: bool = False,
) -> Dict[int, Set[int]]:
    """Add bonds to a fresnel scene.

    Parameters:
        scene: The fresnel scene.
        sites: A zipped list of species and positions.
        bond_dict: The dictionary defining bond lengths between species.
        bond_thickness: The thickness of the bonds in the visualisation.
        nsites: The number of "real" sites in the structure.
        bond_images: Whether to draw bonds between images.

    Returns:
        Indices of bonded atoms as a dictionary of sets.

    """
    import fresnel
    from collections import defaultdict

    bond_colours = []
    bonds = []
    bond_sets = defaultdict(set)
    bond_dict = {} if bond_dict is None else bond_dict
    for k, v in list(bond_dict.items()):
        bond_dict[tuple(sorted(k))] = v

    colours = get_element_colours()
    for i, i_atom in enumerate(sites):
        for j, j_atom in enumerate(sites):
            if j <= i:
                continue
            species = (sites[i][0], sites[j][0])
            dist = np.linalg.norm(np.array(sites[i][1]) - np.array(sites[j][1]))
            if bond_dict.get(tuple(sorted(species)), -1) < dist:
                continue
            # If these are both image atoms
            if not bond_images and nsites and i >= nsites and j >= nsites:
                continue
            bonds.append((sites[i][1], sites[j][1]))
            bond_sets[i].add(j)
            bond_sets[j].add(i)
            bond_colours.append(
                (
                    fresnel.color.linear(colours[sites[i][0]]),
                    fresnel.color.linear(colours[sites[j][0]]),
                )
            )
    if bonds:
        bonds_geom = fresnel.geometry.Cylinder(
            scene,
            N=len(bonds),
            color=bond_colours,
        )
        bonds_geom.points[:] = bonds
        bonds_geom.radius[:] = bond_thickness * np.ones(len(bonds))
        bonds_geom.material.primitive_color_mix = 1
        bonds_geom.outline_width = 0.05

    return bond_sets


def _draw_atoms(
    scene: "fresnel.Scene",
    positions: List[Tuple[float, float, float]],
    species: List[str],
):
    """Add atoms to the scene, coloured by species.

    Parameters:
        scene: The fresnel scene to add the atoms to.
        positions: List of positions of the atoms to draw.
        species: List of species of the atoms to draw.

    Returns:
        atoms: The atoms geometry.
        sites: The list of sites (as species, position tuples) that were added to the scene.

    """
    import fresnel

    colours = get_element_colours()
    radii = get_element_radii()

    atoms = fresnel.geometry.Sphere(
        scene, position=positions, radius=1.0, outline_width=0.1
    )
    atoms.material = fresnel.material.Material()
    atoms.material.solid = 0.0
    atoms.material.primitive_color_mix = 1
    atoms.material.roughness = 0.8
    atoms.material.specular = 0.9
    atoms.material.spec_trans = 0.0
    atoms.material.metal = 0
    atoms.color[:] = fresnel.color.linear([colours[s] for s in species])
    atoms.radius[:] = [0.5 * radii[s] for s in species]


def rerender_scenes_to_axes(scenes, axes, renderer=None):
    """(Re)render the scenes to the axes.

    Parameters:
        scenes: List of fresnel scenes to render.
        axes: List or grid of matplolib axes to render to.
        renderer: The renderer to use (default pathtrace).

    """
    import fresnel

    _axes = axes.flatten() if not isinstance(axes, list) else axes
    if renderer is None:
        from functools import partial

        renderer = partial(fresnel.pathtrace, light_samples=32)

    for i, ax in tqdm.tqdm(enumerate(_axes), desc="Rendering scenes"):
        old_title = ax.title.get_text()
        ax.clear()
        ax.axis("off")
        ax.set_aspect("equal")
        ax.set_title(old_title)
        if i < len(scenes):
            render = renderer(scenes[i])
            ax.imshow(render[:])


def fit_orthographic_camera(
    scene: "fresnel.Scene", scale=1.1, margin=0.05, view="isometric", vectors=None
) -> None:
    """Alternative orthographic camera fitting, based on the equivalent fresnel function.

    Parameters:
        scene: The fresnel scene to visualize.
        scale: The scale of the camera.
        margin: The margin around the viewbox.
        view: Either 'front', 'isometric', 'a', 'b', 'c', 'vector' or an actual vector.
        vectors: A custom dictionary of vectors with keys 'v', 'up' and 'right' that defines
            the camera basis.

    """
    import fresnel

    if not scene.camera:
        scene.camera = fresnel.camera.Orthographic()

    if view == "front":
        view = "c"

    _default_vectors = {
        "a": dict(
            v=np.array([1, 0, 0]), up=np.array([0, 0, 1]), right=np.array([0, 1, 0])
        ),
        "b": dict(
            v=np.array([0, 1, 0]), up=np.array([1, 0, 0]), right=np.array([0, 0, 1])
        ),
        "c": dict(
            v=np.array([0, 0, 1]), up=np.array([0, 1, 0]), right=np.array([1, 0, 0])
        ),
        "isometric": dict(
            v=np.array([1, 1, 1]) / np.sqrt(3),
            up=np.array([-1, 2, -1]) / np.sqrt(6),
            right=np.array([1, 0, -1]) / np.sqrt(2),
        ),
    }

    if view == "vector" or vectors is not None:
        view = "vector"
        _default_vectors["vector"] = vectors
        if "angle" in vectors:
            angle = np.deg2rad(vectors["angle"])
            vectors["v"] = np.array([np.cos(angle), np.sin(angle), 0])
            vectors["right"] = np.array([-np.sin(angle), np.cos(angle), 0])
        if "right" not in vectors:
            vectors["right"] = np.cross(vectors["up"], vectors["v"])
        for k in ("up", "v", "right"):
            vectors[k] = np.array(vectors[k]) / np.linalg.norm(vectors[k])

    if isinstance(view, list):
        _view = np.array(view)
        _view = _view / np.linalg.norm(_view)
        up = [0, 0, 1]
        right = np.cross(_view, up)
        up = np.cross(_view, right)
        vector_view = {"v": _view, "up": up, "right": right}
        _default_vectors["vector"] = vector_view
        view = "vector"

    # raise error if the scene is empty
    if len(scene.geometry) == 0:
        raise ValueError("The scene is empty")

    # find the center of the scene
    extents = scene.get_extents()

    # choose an appropriate view automatically
    if view == "auto":
        xw = extents[1, 0] - extents[0, 0]
        yw = extents[1, 1] - extents[0, 1]
        zw = extents[1, 2] - extents[0, 2]

        if zw < 0.51 * max(xw, yw):
            view = "c"
        else:
            view = "isometric"

    v = _default_vectors[view]["v"]
    up = _default_vectors[view]["up"]

    # make a list of points of the cube surrounding the scene
    points = np.array(
        [
            [extents[0, 0], extents[0, 1], extents[0, 2]],
            [extents[0, 0], extents[0, 1], extents[1, 2]],
            [extents[0, 0], extents[1, 1], extents[0, 2]],
            [extents[0, 0], extents[1, 1], extents[1, 2]],
            [extents[1, 0], extents[0, 1], extents[0, 2]],
            [extents[1, 0], extents[0, 1], extents[1, 2]],
            [extents[1, 0], extents[1, 1], extents[0, 2]],
            [extents[1, 0], extents[1, 1], extents[1, 2]],
        ]
    )

    # find the center of the box
    center = (extents[0, :] + extents[1, :]) / 2
    points = points - center

    # determine the extent of the scene box in the up direction
    up_projection = np.dot(points, up)
    height = (1 + margin) * np.max(np.abs(up_projection)) * 2

    # determine the extent of the scene box in the view direction
    view_projection = np.dot(points, v)
    view_distance = np.max(view_projection) * scale

    # build the camera
    scene.camera.position = center + view_distance * v
    scene.camera.look_at = center
    scene.camera.up = up
    scene.camera.height = height


def get_image_atoms_near_cell_boundaries(
    doc: Crystal, tolerance: float, bond_dict: Dict[Tuple[str, str], float] = None
) -> Tuple[List[str], List[Tuple[float, float, float]]]:
    """Returns a list of atom positions and species that are within the
    distance `tolerance` of the non-image atoms.

    Parameters:
        doc: The crystal object to consider.
        tolerance: The distance tolerance in Å.
        bond_dict: If provided, use asspecies-specific tolerances.

    Returns:
        A list of species and a list of positions in Cartesian coordinates.

    """
    from itertools import product
    from scipy.spatial import KDTree

    n_iter = 1

    tree = KDTree(doc.positions_abs, compact_nodes=False, copy_data=True)

    test_displacements = [
        np.sum(np.array(doc.lattice_cart) * np.array(p), axis=-1)
        for p in product(range(-1, 2), repeat=3)
        if p != (0, 0, 0)
    ]
    positions = []
    species = []
    added = set()
    for j, pos in enumerate(doc.positions_abs):
        for d, disp in enumerate(test_displacements):
            image_pos = pos + disp
            if tree.query_ball_point(image_pos, tolerance * 1.2):
                positions.append(image_pos.tolist())
                species.append(doc.atom_types[j])
                added.add((j, d))

    # Can also "self-consistently" check for images close to the just-added images
    # Not sure this is worth enabling
    if n_iter == 2:
        tree = KDTree(positions, compact_nodes=False, copy_data=True)
        for j, pos in enumerate(doc.positions_abs):
            for d, disp in enumerate(test_displacements):
                if (j, d) not in added:
                    image_pos = pos + disp
                    if tree.query_ball_point(image_pos, tolerance):
                        positions.append(image_pos.tolist())
                        species.append(doc.atom_types[j])

    return positions, species


def fresnel_plot(
    structures: Union[Crystal, List[Crystal]],
    figsize: Optional[Tuple[float, float]] = None,
    fig_cols: Optional[int] = None,
    fig_rows: Optional[int] = None,
    labels: Union[bool, List[str]] = True,
    renderer: Optional[Callable] = None,
    camera_patches: Optional[List[Optional[Dict]]] = None,
    **fresnel_view_kwargs,
):
    """Visualize a series of structures as a grid of matplotlib plots.

    Parameters:
        structures: A list or single structure to visualize.
        figsize: The size of the figure (defaults to a 3x4 box for each structure).
        fig_cols: The number of columns in the figure.
        fig_rows: The number of rows in the figure.
        labels: If true, label structures by formula, otherwise label by the passed
            list of strings.
        renderer: A callable to use instead of the default `fresnel.pathtrace`.
            A useful alternative could be `fresnel.preview`. Can also be used to
            parameterise the fresnel rendering with e.g.,
            `partial(fresnel.pathtrace, samples=64)`.
        fresnel_view_kwargs: Any additional kwargs will be passed down to
            `matador.utils.viz_utils.fresnel_view` to control e.g., camera angles.
            The contents of any keywords containing lists with the same length as
            the passed structures will be applied to each structure individually.

    Returns:
        The matplotlib figures, axis grid and a list of fresnel scenes.

    """
    import matplotlib.pyplot as plt

    if isinstance(structures, Crystal):
        structures = [structures]

    if fig_cols is None and fig_rows is None:
        fig_cols = len(structures)
        fig_rows = 1
    elif fig_rows is None:
        fig_rows = -(len(structures) // -fig_cols)
    else:
        fig_cols = -(len(structures) // -fig_rows)

    if figsize is None:
        figsize = (2 * fig_cols, 3 * fig_rows)

    fig, axes = plt.subplots(
        nrows=fig_rows, ncols=fig_cols, figsize=figsize, squeeze=False
    )

    scenes = []

    list_kwargs = []
    for kw in fresnel_view_kwargs:
        if isinstance(fresnel_view_kwargs[kw], list) or isinstance(
            fresnel_view_kwargs[kw], tuple
        ):
            if len(fresnel_view_kwargs[kw]) == len(structures):
                list_kwargs.append(kw)

    for ind, s in enumerate(structures):
        kwargs = {k: fresnel_view_kwargs[k][ind] for k in list_kwargs}
        kwargs.update(
            {
                k: fresnel_view_kwargs[k]
                for k in fresnel_view_kwargs
                if k not in list_kwargs
            }
        )
        scenes.append(fresnel_view(s, **kwargs))

    if camera_patches is not None:
        for ind, scene in enumerate(scenes):
            patch = camera_patches[ind] or {}
            if "view" in patch:
                fit_orthographic_camera(scene, view=patch["view"])
            if "height" in patch:
                scene.camera.height = scene.camera.height + patch["height"]

    if labels is not None:
        if isinstance(labels, bool) and labels:
            labels = [s.formula_tex for s in structures]
        if labels:
            for ind, ax in enumerate(axes.flatten()):
                try:
                    ax.set_title(labels[ind])
                except IndexError:
                    pass

    rerender_scenes_to_axes(scenes, axes, renderer=renderer)

    fig.set_tight_layout(True)

    return fig, axes, scenes


def formula_to_colour(formula: str) -> List[float]:
    """Return an RGBA colour for the given chemical formula, provided
    the formula has 3 or fewer species.

    """
    from matador.utils.chem_utils import get_stoich_from_formula, get_concentration

    stoichiometry = get_stoich_from_formula(formula)
    elements = [d[0] for d in stoichiometry]
    if len(elements) > 3:
        raise RuntimeError(
            f"Cannot mix a colour for more than 3 elements, received: {stoichiometry}"
        )

    concentration = get_concentration(
        stoichiometry, elements=elements, include_end=True
    )

    return colour_from_ternary_concentration(concentration, species=elements)


def colour_from_ternary_concentration(
    conc: Union[Tuple[float, float], Tuple[float, float, float]],
    species: List[str],
    alpha: float = 1,
) -> List[float]:
    """Returns a colour for a ternary composition of the given species, mixing
    the configured VESTA colours.

    Returns:
        RGBA array.

    """
    if len(conc) == 1:
        return get_element_colours()[species[0]] + [alpha]
    elif len(conc) == 2:
        x, y = conc
        z = 1 - x - y
    else:
        x, y, z = conc
    colours = [np.asarray(get_element_colours()[s]) for s in species + ["H"]]
    return ((x * colours[0] + y * colours[1] + z * colours[2])).tolist() + [alpha]



# coding: utf-8
# Distributed under the terms of the MIT License.

""" This file implements some scraper functions for
Quantum Espresso-related inputs and outputs.

"""


from collections import defaultdict
from os import stat
from math import gcd
from matador.utils.cell_utils import cart2abc, cart2volume
from matador.utils.chem_utils import RY_TO_EV, KBAR_TO_GPA
from matador.scrapers.utils import scraper_function, get_flines_extension_agnostic


@scraper_function
def pwout2dict(fname, **kwargs):
    """Extract available information from pw.x .out file.

    Parameters:
        fname (str/list): filename or list of filenames to scrape as a
            QuantumEspresso pw.x output.

    """
    flines, fname = get_flines_extension_agnostic(fname, ["out", "in"])

    pwout = {}
    pwout["source"] = [fname]

    try:
        # grab file owner username
        from pwd import getpwuid

        pwout["user"] = getpwuid(stat(fname).st_uid).pw_name
    except Exception:
        pwout["user"] = "xxx"

    if "CollCode" in fname:
        pwout["icsd"] = fname.split("CollCode")[-1]
    for ind, line in enumerate(reversed(flines)):
        ind = len(flines) - 1 - ind
        if (
            "cell_parameters" in line.lower()
            and "angstrom" in line.lower()
            and "lattice_cart" not in pwout
        ):
            pwout["lattice_cart"] = []
            for j in range(3):
                line = flines[ind + j + 1].strip().split()
                pwout["lattice_cart"].append(list(map(float, line)))
            pwout["cell_volume"] = cart2volume(pwout["lattice_cart"])
        elif "atomic_positions" in line.lower() and "positions_frac" not in pwout:
            pwout["positions_frac"] = []
            pwout["atom_types"] = []
            j = 1
            while True:
                if "End final coordinates" in flines[j + ind]:
                    break
                else:
                    try:
                        line = flines[j + ind].strip().split()
                        pwout["atom_types"].append(line[0])
                        pwout["positions_frac"].append(list(map(float, line[1:5])))
                        j += 1
                    except Exception:
                        break
            pwout["num_atoms"] = len(pwout["atom_types"])
        elif "final enthalpy" in line.lower() and "enthalpy" not in pwout:
            pwout["enthalpy"] = RY_TO_EV * float(line.lower().split()[-2])
        elif "total   stress" in line.lower() and "pressure" not in pwout:
            pwout["pressure"] = KBAR_TO_GPA * float(line.lower().split()[-1])
        elif all(
            key in pwout
            for key in ["enthalpy", "pressure", "lattice_cart", "positions_frac"]
        ):
            break
    # get abc lattice
    pwout["lattice_abc"] = cart2abc(pwout["lattice_cart"])
    # calculate stoichiometry
    pwout["stoichiometry"] = defaultdict(float)
    for atom in pwout["atom_types"]:
        if atom not in pwout["stoichiometry"]:
            pwout["stoichiometry"][atom] = 0
        pwout["stoichiometry"][atom] += 1
    gcd_val = 0
    for atom in pwout["atom_types"]:
        if gcd_val == 0:
            gcd_val = pwout["stoichiometry"][atom]
        else:
            gcd_val = gcd(pwout["stoichiometry"][atom], gcd_val)
    # convert stoichiometry to tuple for fryan
    temp_stoich = []
    for key, value in pwout["stoichiometry"].items():
        if float(value) / gcd_val % 1 != 0:
            temp_stoich.append([key, float(value) / gcd_val])
        else:
            temp_stoich.append([key, value / gcd_val])
    pwout["stoichiometry"] = temp_stoich
    atoms_per_fu = 0
    for elem in pwout["stoichiometry"]:
        atoms_per_fu += elem[1]
    pwout["num_fu"] = len(pwout["atom_types"]) / atoms_per_fu

    return pwout, True



# coding: utf-8
# Distributed under the terms of the MIT License.

""" This submodule contains functions to plot convergence test data
for a series of calculations.

"""

import glob
from decimal import Decimal, ROUND_UP
from collections import defaultdict
from traceback import print_exc
import numpy as np
from matador.plotting.plotting import plotting_function, SAVE_EXTS
from matador.scrapers.castep_scrapers import castep2dict
from matador.utils.chem_utils import get_formula_from_stoich

__all__ = ["plot_cutoff_kpt_grid"]


@plotting_function
def plot_cutoff_kpt_grid(
    data,
    figsize=None,
    forces=False,
    max_energy=None,
    max_force=None,
    legend=True,
    colour_by="formula",
    log=False,
    **kwargs,
):
    """Create a composite plot of cutoff energy and kpoint convergence.

    Parameters:
        data (dict): dictionary of convergence data.

    Keyword arguments:
        forces (bool): whether or not to plot forces.
        legend (bool): whether or not to show a legend.
        max_energy (float): max ylimit for convergence data in meV.

    """
    import matplotlib.pyplot as plt

    if figsize is None:
        if forces:
            figsize = (8, 7)
        else:
            figsize = (8, 4)

    if forces:
        _, axes = plt.subplots(2, 2, figsize=figsize, sharex="col", sharey="row")
    else:
        _, axes = plt.subplots(1, 2, figsize=figsize, sharey="row")

    if forces:
        ax = axes[0][0]
    else:
        ax = axes[0]

    k_label = "Reduced $k$-point spacing (1/Å)"
    e_label = "Planewave cutoff energy (eV)"

    _, lines, labels = plot_field(
        data,
        field="formation_energy_per_atom",
        parameter="cut_off_energy",
        ax=ax,
        colour_by=colour_by,
        max_y=max_energy,
        log=log,
        label_x=e_label if not forces else False,
        label_y=True,
        reference="last",
    )

    if len(labels) < 30 and legend:
        plt.figlegend(lines, labels, loc="upper center", ncol=min(len(lines) // 2, 8))

    if forces:
        ax = axes[0][1]
    else:
        ax = axes[1]

    plot_field(
        data,
        field="formation_energy_per_atom",
        parameter="kpoints_mp_spacing",
        ax=ax,
        colour_by=colour_by,
        max_y=max_energy,
        log=log,
        label_x=k_label if not forces else False,
        label_y=False,
        reference="last",
    )

    if forces:
        plot_field(
            data,
            field="forces",
            parameter="cut_off_energy",
            ax=axes[1][0],
            max_y=max_force,
            label_x=e_label,
            log=log,
            label_y=True,
            reference="last",
        )
        plot_field(
            data,
            field="forces",
            parameter="kpoints_mp_spacing",
            max_y=max_force,
            log=log,
            ax=axes[1][1],
            label_x=k_label,
            label_y=False,
            reference="last",
        )

    plt.tight_layout()

    if any(kwargs.get(ext) for ext in SAVE_EXTS):
        fname = "conv"
        for ext in SAVE_EXTS:
            if kwargs.get(ext):
                plt.savefig(
                    "{}.{}".format(fname, ext), bbox_inches="tight", transparent=False
                )
                print("Wrote {}.{}".format(fname, ext))

    if kwargs.get("plot_fname") or any([kwargs.get(ext) for ext in SAVE_EXTS]):
        import os

        fname = kwargs.get("plot_fname") or "conv"
        for ext in SAVE_EXTS:
            if kwargs.get(ext):
                fname_tmp = fname
                ind = 0
                while os.path.isfile("{}.{}".format(fname_tmp, ext)):
                    ind += 1
                    fname_tmp = fname + str(ind)

                fname = fname_tmp
                plt.savefig(
                    "{}.{}".format(fname, ext), bbox_inches="tight", transparent=True
                )
                print("Wrote {}.{}".format(fname, ext))

    if kwargs.get("show"):
        plt.show()


def plot_field(
    data,
    field="formation_energy_per_atom",
    parameter="cut_off_energy",
    ax=None,
    reference="last",
    max_y=None,
    log=False,
    colour_by="formula",
    label_x=True,
    label_y=True,
    plot_points=True,
):
    """Plot the convergence fields for each structure in `data` at
    each value of `parameter`.

    Parameters:
        data (dict): nested dictionary with keys for each structure
            that stores the convergence data to be plotted. e.g.
            {'structure_A':
                {'formation_energy_per_atom': [1, 2, 3],
                 'cutoff': [500, 400, 300]
                }
            }.

    Keyword arguments:
        field (str): the key under which convergence data is stored. This
            string is used for the y-axis label.
        parameter (str): the key under which the convergence parameter
            values are stored. This string is used for the x-axis label.
        ax (matplotlib.Axis): optional axis object to use for plot.
        reference (str): if 'last' then use the last value of the
            convergence data as the zero.
        plot_points (bool): whether or not to show markers

    """
    import matplotlib.pyplot as plt

    if ax is None:
        fig, ax = plt.subplots()

    lines = []
    labels = []
    labelled = []

    colourmap = {}
    if colour_by == "formula":
        formulae = sorted(
            list(
                {
                    get_formula_from_stoich(data[key]["stoichiometry"], tex=True)
                    for key in data
                }
            )
        )
        for formula in formulae:
            colourmap[formula] = next(ax._get_lines.prop_cycler)["color"]

    for ind, key in enumerate(data):

        formula = get_formula_from_stoich(data[key]["stoichiometry"], tex=True)
        try:
            values, parameters = get_convergence_values(
                data[key], parameter, field, reference=reference, log=log
            )
        except Exception:
            print("Missing data for {}->{}->{}".format(key, parameter, field))
            continue

        label = None
        if colour_by == "formula":
            c = colourmap[formula]
            if formula not in labelled:
                label = formula
                labelled.append(formula)
        else:
            c = next(ax._get_lines.prop_cycler)["color"]
            label = key

        # only plot points if there's less than 25 lines
        if plot_points:
            ax.plot(
                parameters,
                values,
                "o",
                markersize=5,
                alpha=1,
                label=label,
                lw=0,
                zorder=1000,
                c=c,
            )
            ax.plot(parameters, values, "-", alpha=0.2, lw=1, zorder=1000, c=c)
        else:
            ax.plot(
                parameters, values, "-", alpha=0.5, lw=1, zorder=1000, label=label, c=c
            )

        if label is not None:
            lines, labels = ax.get_legend_handles_labels()

    if label_x:
        if isinstance(label_x, bool):
            label_x = parameter.replace("_", " ")
        ax.set_xlabel(label_x)

    if label_y:
        if "force" in field.lower():
            ylabel = "$\\Delta F$ (eV/Å)"
        elif "formation" in field.lower():
            ylabel = "$\\Delta E$ (meV/atom)"
        else:
            ylabel = field.replace("_", " ")

        ax.set_ylabel(ylabel)

    if max_y is not None:
        ax.set_ylim(None, max_y)

    return ax, lines, labels


def round(n, prec):
    """Replace default (bankers) rounding with "normal" rounding."""
    if prec is None:
        return n
    else:
        return float(Decimal(str(n)).quantize(Decimal("0.05"), rounding=ROUND_UP))


def get_convergence_files(path, only=None):
    """Find all CASTEP files in the directory."""
    structure_files = defaultdict(list)
    files = glob.glob(path + "/*.castep")
    for file in files:
        if only is None or only in file:
            castep_dict, success = castep2dict(file, db=False)
            if not success:
                print("Failure to read castep file {}".format(file))
            else:
                source = castep_dict["source"][0].split("/")[-1]
                source = source.replace(".castep", "")
                source = "".join(source.split("_")[:-1])
                structure_files[source].append(castep_dict)

    return structure_files


def get_convergence_data(
    structure_files, conv_parameter="cut_off_energy", species=None
):
    """Parse cutoff energy/kpt spacing convergence calculations from list of files.

    Parameters:
        structure_files (list): list of filenames.

    Keyword arguments:
        conv_parameter (str): field for convergence parameter.
        species (list): only include structures with these species included.

    """
    scraped_from_filename = None

    form_set = {
        get_formula_from_stoich(
            structure_files[structure][0]["stoichiometry"], tex=True
        )
        for structure in structure_files
    }
    if len(form_set) == 1:
        chempot_mode = False
        single = True

    else:
        chempot_mode = True
        single = False
        chempots_dict = defaultdict(dict)

    if conv_parameter == "kpoints_mp_spacing":
        rounding = None
    else:
        rounding = None

    data = {}

    if chempot_mode:
        for key in structure_files:
            skip = False
            if species is not None:
                for elem in structure_files[key][0]["stoichiometry"]:
                    if elem[0] not in species:
                        skip = True
            if skip:
                continue
            for doc in structure_files[key]:
                scraped_from_filename = None
                if not single and len(doc["stoichiometry"]) == 1:
                    if conv_parameter == "kpoints_mp_spacing":
                        try:
                            scraped_from_filename = float(
                                doc["source"][0]
                                .split("/")[-1]
                                .split("_")[-1]
                                .split("A")[0]
                            )
                        except ValueError:
                            print(
                                f"Unable to determine kpoints label from {doc['source'][0]}, skipping..."
                            )
                            continue

                    if scraped_from_filename is not None:
                        rounded_field = round(scraped_from_filename, rounding)
                    else:
                        rounded_field = round(doc[conv_parameter], rounding)
                    energy_key = "total_energy_per_atom"
                    chempots_dict[str(rounded_field)][doc["atom_types"][0]] = doc[
                        energy_key
                    ]

    elems = set()

    if chempot_mode:
        for value in chempots_dict:
            for elem in chempots_dict[value]:
                elems.add(elem)

        for value in chempots_dict:
            for elem in elems:
                if elem not in chempots_dict[value]:
                    print(
                        "WARNING: {} chemical potential missing at {} = {} eV, skipping this value.".format(
                            elem, conv_parameter, value
                        )
                    )

    for key in structure_files:
        skip = False
        if species is not None:
            for elem in structure_files[key][0]["stoichiometry"]:
                if elem[0] not in species:
                    skip = True
        if skip:
            continue

        data[key] = {}
        data[key][conv_parameter] = defaultdict(list)

        for doc in structure_files[key]:
            if conv_parameter == "kpoints_mp_spacing":
                try:
                    scraped_from_filename = float(
                        doc["source"][0].split("/")[-1].split("_")[-1].split("A")[0]
                    )
                except ValueError:
                    print(
                        f"Unable to determine kpoints label from {doc['source'][0]}, skipping..."
                    )
                    continue
            try:
                doc["formation_energy_per_atom"] = doc["total_energy_per_atom"]
                if scraped_from_filename is not None:
                    rounded_field = round(scraped_from_filename, rounding)
                else:
                    rounded_field = round(doc[conv_parameter], rounding)
                if chempot_mode and len(doc["stoichiometry"]) > 1:
                    for atom in doc["atom_types"]:
                        doc["formation_energy_per_atom"] -= chempots_dict[
                            str(rounded_field)
                        ][atom] / len(doc["atom_types"])

                data[key][conv_parameter][conv_parameter].append(rounded_field)
                data[key]["stoichiometry"] = doc["stoichiometry"]
                data[key][conv_parameter]["formation_energy_per_atom"].append(
                    doc["formation_energy_per_atom"]
                )

                if "forces" in doc:
                    data[key][conv_parameter]["forces"].append(
                        np.sqrt(np.sum(np.asarray(doc["forces"]) ** 2, axis=-1))
                    )

            except Exception:
                print_exc()
                print("Error with {}".format(key))

    if conv_parameter == "kpoints_mp_spacing":
        reverse = True
    else:
        reverse = False

    for key in data:
        inds = [
            ind
            for ind, _ in sorted(
                enumerate(data[key][conv_parameter][conv_parameter]),
                key=lambda x: x[1],
                reverse=reverse,
            )
        ]
        for field in data[key][conv_parameter]:
            if field != "stoichiometry":
                data[key][conv_parameter][field] = [
                    data[key][conv_parameter][field][ind] for ind in inds
                ]

    return data


def get_convergence_values(data, parameter, field, reference="last", log=False):
    """Extract the data to plot for the given dictionary."""

    values = data[parameter][field]
    parameters = data[parameter][parameter]
    if len(values) != len(parameters):
        raise RuntimeError(
            "Mismatched convergence data for {}->{}: {}".format(field, parameter, data)
        )

    values = np.asarray(values)
    parameters = np.asarray(parameters)

    if reference == "last":
        values -= values[-1]
        if "force" not in field:
            values *= 1000
        values = np.abs(values)

    if log:
        with np.errstate(divide="ignore"):
            values = np.log10(values)

    return values, parameters


def combine_convergence_data(data_A, data_B):
    """Combine dictionaries with potentially overlapping keys."""
    data = {}
    for key in data_A:
        data[key] = data_A[key]
        if key in data_B:
            data[key].update(data_B[key])
    for key in data_B:
        if key not in data:
            data[key] = data_B[key]

    return data



# coding: utf-8
# Distributed under the terms of the MIT License.

""" The scrapers module implements the construction of Python dicts from
many common DFT/crystallography file formats.

Currently supported filetypes:

    - CASTEP filetypes: .cell, .param, .castep, .bands, .phonon, .usp
    - Custom .res type based on SHELX files, first used by AIRSS package https://www.mtg.msm.cam.ac.uk/Codes/AIRSS
    - OptaDOS output: .odo, .adaptive.dat
    - Quantum Espresso output files: .out
    - Magres files: .magres
    - Crystallographic Information File (via ASE): .cif

"""


__all__ = [
    "res2dict",
    "cell2dict",
    "param2dict",
    "castep2dict",
    "bands2dict",
    "arbitrary2dict",
    "phonon2dict",
    "phonon_dos2dict",
    "optados2dict",
    "usp2dict",
    "pwout2dict",
    "magres2dict",
    "cif2dict",
]
__author__ = "Matthew Evans"
__maintainer__ = "Matthew Evans"

from .castep_scrapers import res2dict, cell2dict, param2dict
from .castep_scrapers import castep2dict, bands2dict, arbitrary2dict
from .castep_scrapers import phonon2dict, phonon_dos2dict, optados2dict, usp2dict
from .qe_scrapers import pwout2dict
from .magres_scrapers import magres2dict
from .cif_scraper import cif2dict



# coding: utf-8
# Distributed under the terms of the MIT License.

""" The workflows.castep submodule contains any workflows that rely on
CASTEP-specific implementation, for example electronic spectropscopy or phonons.

"""


__all__ = (
    "castep_full_phonon",
    "castep_full_spectral",
    "castep_elastic",
    "castep_full_magres",
)

__author__ = "Matthew Evans"
__maintainer__ = "Matthew Evans"


from .phonons import castep_full_phonon
from .spectral import castep_full_spectral
from .elastic import castep_elastic
from .magres import castep_full_magres



# coding: utf-8
# Distributed under the terms of the MIT License.

""" This file implements functions that write files from matador
documents or Crystal objects.
"""

import os
import warnings

import numpy as np
import pymongo as pm

from matador.utils.cell_utils import cart2abcstar, frac2cart, cart2abc
from matador.utils.cell_utils import abc2cart, calc_mp_grid
from matador.utils.cursor_utils import display_results
from matador.utils.chem_utils import (
    get_formula_from_stoich,
    get_stoich,
    get_root_source,
)
from matador.swaps import AtomicSwapper
from .utils import file_writer_function, generate_relevant_path, generate_hash

EPS = 1e-8


def query2files(
    cursor,
    dirname=None,
    max_files=10000,
    top=None,
    prefix=None,
    cell=None,
    param=None,
    res=None,
    pdb=None,
    json=None,
    xsf=None,
    markdown=True,
    latex=False,
    subcmd=None,
    argstr=None,
    **kwargs
):
    """Many-to-many convenience function for many structures being written to
    many file types.

    Parameters:
        cursor (:obj:`list` of :obj:`dict`/:class:`AtomicSwapper`): list of matador dictionaries to write out.

    Keyword arguments:
        dirname (str): the folder to save the results into. Will be created if non-existent.
            Will have integer appended to it if already existing.
        max_files (int): if the number of files to be written exceeds this number, then raise RuntimeError.
        **kwargs (dict): dictionary of {filetype: bool(whether to write)}. Accepted file types
            are cell, param, res, pdb, json, xsf, markdown and latex.

    """
    multiple_files = any((cell, param, res, pdb, xsf))
    prefix = prefix + "-" if prefix is not None else ""

    if isinstance(cursor, AtomicSwapper):
        cursor = cursor.cursor
        subcmd = "swaps"

    if subcmd in ["polish", "swaps"]:
        info = False
        hash_dupe = False
    else:
        info = True
        hash_dupe = False

    if isinstance(cursor, list):
        num = len(cursor)
    else:
        num = cursor.count()

    if top is not None:
        if top < num:
            num = top

    num_files = num * sum(1 for ext in [cell, param, res, pdb, xsf] if ext)

    if multiple_files:
        if num_files > max_files:
            raise RuntimeError(
                "Not writing {} files as it exceeds argument `max_files` limit of {}".format(
                    num_files, max_files
                )
            )

    if dirname is None:
        dirname = generate_relevant_path(subcmd=subcmd, **kwargs)

    _dir = False
    dir_counter = 0
    # postfix integer on end of directory name if it exists
    while not _dir:
        if dir_counter != 0:
            directory = dirname + str(dir_counter)
        else:
            directory = dirname
        if not os.path.isdir(directory):
            os.makedirs(directory)
            _dir = True
        else:
            dir_counter += 1

    for _, doc in enumerate(cursor[:num]):
        # generate an appropriate filename for the structure
        root_source = get_root_source(doc)

        if "_swapped_stoichiometry" in doc:
            formula = get_formula_from_stoich(doc["_swapped_stoichiometry"])
        else:
            formula = get_formula_from_stoich(doc["stoichiometry"])

        if subcmd == "swaps":
            root_source = root_source.replace("-swap-", "-")

        name = root_source

        if "OQMD " in root_source:
            name = "{formula}-OQMD_{src}".format(
                formula=formula, src=root_source.split(" ")[-1]
            )
        elif "mp-" in root_source:
            name = "{formula}-MP_{src}".format(
                formula=formula, src=root_source.split("-")[-1]
            )
        if "icsd" in doc and "CollCode" not in name:
            name += "-CollCode{}".format(doc["icsd"])
        else:
            pf_id = None
            for source in doc["source"]:
                if "pf-" in source:
                    pf_id = source.split("-")[-1]
                    break
            else:
                if "pf_ids" in doc:
                    pf_id = doc["pf_ids"][0]
            if pf_id is not None:
                name += "-PF-{}".format(pf_id)

        # if swaps, prepend new composition
        if subcmd == "swaps":
            new_formula = get_formula_from_stoich(get_stoich(doc["atom_types"]))
            name = "{}-swap-{}".format(new_formula, name)

        path = "{directory}/{prefix}{name}".format(
            directory=directory, prefix=prefix, name=name
        )

        if param:
            doc2param(doc, path, hash_dupe=hash_dupe)
        if cell:
            doc2cell(doc, path, hash_dupe=hash_dupe)
        if res:
            doc2res(doc, path, info=info, hash_dupe=hash_dupe)
        if json:
            doc2json(doc, path, hash_dupe=hash_dupe)
        if pdb:
            doc2pdb(doc, path, hash_dupe=hash_dupe)
        if xsf:
            doc2xsf(doc, path)

    hull = subcmd in ["hull", "voltage"]
    if isinstance(cursor, pm.cursor.Cursor):
        cursor.rewind()
    md_path = "{directory}/{directory}.md".format(directory=directory)
    md_kwargs = {}
    md_kwargs.update(kwargs)
    md_kwargs.update({"markdown": True, "latex": False, "argstr": argstr, "hull": hull})
    md_string = display_results(cursor, **md_kwargs)
    with open(md_path, "w") as f:
        f.write(md_string)

    if latex:
        if isinstance(cursor, pm.cursor.Cursor):
            cursor.rewind()
        tex_path = "{directory}/{directory}.tex".format(directory=directory)
        tex_kwargs = {}
        tex_kwargs.update(kwargs)
        tex_kwargs.update(
            {"latex": True, "markdown": False, "argstr": argstr, "hull": hull}
        )
        tex_string = display_results(cursor, **tex_kwargs)
        with open(tex_path, "w") as f:
            f.write(tex_string)


@file_writer_function
def doc2param(doc, path, overwrite=False, hash_dupe=False, spin=False):
    """Write basic .param file from single doc.

    Parameters:
        doc (dict): the document to write out to file
        path (str): the desired path to file

    Keyword arguments:
        spin (bool): enforce breaking of spin symmetry to magic number of 5.
        overwrite (bool): whether or not to overwrite colliding files.
        hash_dupe (bool): whether or not to create a unique filename for
            any colliding files, or just skip writing them.

    Returns:
        list, str: list of strings to write to file, with implicit newlines,
            and the required file extension.

    """
    from matador.utils.castep_params import CASTEP_PARAMS, CASTEP_VERSION

    param_set = set(CASTEP_PARAMS)
    param_dict = dict()
    for param in [param for param in param_set if param in doc]:
        param_dict[param] = doc[param]

    flines = []
    flines.append(
        "# Param file generated by matador (Matthew Evans 2016)\n# CASTEP version {} parameter set".format(
            CASTEP_VERSION
        )
    )

    if isinstance(spin, bool):
        if spin:
            spin = 5
        else:
            spin = None

    skip_keywords = ["nbands", "nelectrons"]
    for kw in skip_keywords:
        if kw in doc:
            param_dict.pop(kw)

    if spin is not None and not doc.get("spin_polarized"):
        param_dict["spin_polarized"] = True

    if param_dict.get("spin_polarized"):
        total_spin = None
        if "atomic_init_spins" in doc:
            total_spin = sum([val for val in doc["atomic_init_spins"] if val])
        elif spin is not None:
            total_spin = spin

        if total_spin is not None:
            flines.append("{0:30}: {1}".format("spin", total_spin))

    for param in param_dict:
        if param not in ["source", "devel_code"]:
            if param in ["basis_precision"] and "cut_off_energy" in param_dict:
                continue
            flines.append("{0:30}: {1}".format(param, param_dict[param]))

    if "encapsulated" in doc and doc["encapsulated"]:
        try:
            from implicit_cnts import implicit_cnt_params

            cnt_params = implicit_cnt_params(doc["cnt_radius"])
            flines.append("%BLOCK DEVEL_CODE")
            flines.append('ADD_EXT_LOCPOT: "gaussian_cylinder"')
            flines.append("1D_STRESS_TENSOR: True")
            flines.append("gaussian_cylinder_pot:")
            flines.append(
                'V0 = {V0}\nradius = {radius}\nbroadening = {fwhm}\naxial = "0 0 1"\ncentre = "0.5 0.5 0"\n'.format(
                    **cnt_params
                )
            )
            flines.append(":endgaussian_cylinder_pot")
            flines.append("%ENDBLOCK DEVEL_CODE")
        except ImportError as exc:
            raise ImportError(
                "Failed to import implicit_cnt_params from pyairss, so not writing "
                "implicit cnt parameters from .res file to .param file."
            ) from exc

    if "devel_code" in param_dict:
        flines.append("\n%BLOCK DEVEL_CODE")
        flines.append(param_dict["devel_code"].strip())
        flines.append("%ENDBLOCK DEVEL_CODE")

    ext = "param"

    return flines, ext


@file_writer_function
def doc2cell(doc, path, overwrite=False, hash_dupe=False, spin=False):
    """Write .cell file for single doc.

    Parameters:
        doc (dict): the document to write to file
        path (str): the desired path to the file

    Keyword Arguments:
        overwrite (bool): whether or not to overwrite colliding files.
        hash_dupe (bool): whether or not to create a unique filename for
            any colliding files, or just skip writing them.
        spin (bool): break spin symmetry with magic number 5 on first atom

    Returns:
        list, str: list of strings to write to file, with implicit newlines,
            and the required file extension.

    """
    if isinstance(spin, bool):
        if spin:
            spin = 5
        else:
            spin = None

    # if spin symmetry breaking is requested, update atomic init spins
    if "positions_frac" in doc and len(doc["positions_frac"]) > 0:
        if spin or "atomic_init_spins" in doc:
            if not any(doc.get("atomic_init_spins", [])):
                doc["atomic_init_spins"] = len(doc["positions_frac"]) * [None]
                doc["atomic_init_spins"][0] = spin

    if "site_occupancy" in doc:
        for occ in doc["site_occupancy"]:
            if abs(occ - 1) > EPS:
                raise RuntimeError(
                    "Partial occupancies (VCA) not supported in cell files with matador"
                )

    flines = []
    flines.append("# Cell file generated by matador (Matthew Evans 2016)\n")

    if "lattice_cart" in doc:
        flines.append("%BLOCK LATTICE_CART")
        for vec in doc["lattice_cart"]:
            flines.append("{d[0]} {d[1]} {d[2]}".format(d=vec))
        flines.append("%ENDBLOCK LATTICE_CART")

    if "atom_types" and ("positions_frac" in doc or "positions_abs" in doc):

        if "positions_frac" in doc:
            title = "POSITIONS_FRAC"
            positions = doc["positions_frac"]
        else:
            title = "POSITIONS_ABS"
            positions = doc["positions_abs"]

        flines.append("\n%BLOCK {}".format(title))
        for ind, atom in enumerate(zip(doc["atom_types"], positions)):
            postfix = ""
            try:
                if "atomic_init_spins" in doc and doc["atomic_init_spins"][ind]:
                    postfix = "SPIN={}".format(doc["atomic_init_spins"][ind])
            except IndexError:
                pass

            flines.append(
                "{0:8s} {1[0]: 20.15f} {1[1]: 20.15f} {1[2]: 20.15f} {2:}".format(
                    atom[0], atom[1], postfix
                )
            )

        flines.append("%ENDBLOCK {}".format(title))

    if "external_pressure" in doc:
        flines.append("\n%BLOCK EXTERNAL_PRESSURE")
        if np.asarray(doc["external_pressure"]).shape == (3, 3):
            flines.append(
                "{d[0][0]} {d[0][1]} {d[0][2]}".format(d=doc["external_pressure"])
            )
            flines.append("{d[1][1]} {d[1][2]}".format(d=doc["external_pressure"]))
            flines.append("{d[2][2]}".format(d=doc["external_pressure"]))
        else:
            flines.append(
                "{d[0][0]} {d[0][1]} {d[0][2]}".format(d=doc["external_pressure"])
            )
            flines.append("{d[1][0]} {d[1][1]}".format(d=doc["external_pressure"]))
            flines.append("{d[2][0]}".format(d=doc["external_pressure"]))

        flines.append("%ENDBLOCK EXTERNAL_PRESSURE")

    if "external_efield" in doc:
        flines.append("\n%BLOCK EXTERNAL_EFIELD")
        flines.append("{d[0]} {d[1]} {d[2]}".format(d=doc["external_efield"]))
        flines.append("%ENDBLOCK EXTERNAL_EFIELD")

    if "ionic_constraints" in doc:
        flines.append("\n%BLOCK IONIC_CONSTRAINTS")
        for constraint in doc["ionic_constraints"]:
            flines.append("{}".format(constraint))
        flines.append("%ENDBLOCK IONIC_CONSTRAINTS")

    # specify SCF kpoints
    if "kpoints_list" in doc:
        flines.append("\n%BLOCK KPOINTS_LIST")
        for kpoint in doc["kpoints_list"]:
            flines.append("{p[0]:f} {p[1]:f} {p[2]:f} {p[3]:f}".format(p=kpoint))
        flines.append("\n%ENDBLOCK KPOINTS_LIST")
    elif "kpoints_mp_spacing" in doc:
        flines.append("\nKPOINTS_MP_SPACING: {}".format(doc["kpoints_mp_spacing"]))
    elif "kpoints_mp_grid" in doc:
        flines.append(
            "\nKPOINTS_MP_GRID: {d[0]} {d[1]} {d[2]}".format(d=doc["kpoints_mp_grid"])
        )

    if "supercell_kpoints_mp_spacing" in doc:
        flines.append(
            "\nSUPERCELL_KPOINTS_MP_SPACING: {}".format(
                doc["supercell_kpoints_mp_spacing"]
            )
        )

    if "kpoints_mp_offset" in doc:
        flines.append(
            "\nKPOINTS_MP_OFFSET: {d[0]} {d[1]} {d[2]}".format(
                d=doc["kpoints_mp_offset"]
            )
        )

    if "spectral_kpoints_mp_offset" in doc:
        flines.append(
            "SPECTRAL_KPOINTS_MP_OFFSET: {d[0]} {d[1]} {d[2]}".format(
                d=doc["spectral_kpoints_mp_offset"]
            )
        )
    if "phonon_kpoint_mp_offset" in doc:
        flines.append(
            "PHONON_KPOINT_MP_OFFSET: {d[0]} {d[1]} {d[2]}".format(
                d=doc["phonon_kpoint_mp_offset"]
            )
        )

    # specify spectral kpoints
    if "spectral_kpoints_mp_spacing" in doc:
        flines.append(
            "\nSPECTRAL_KPOINTS_MP_SPACING: {}".format(
                doc["spectral_kpoints_mp_spacing"]
            )
        )
    elif "spectral_kpoints_mp_grid" in doc:
        flines.append(
            "\nSPECTRAL_KPOINTS_MP_GRID: {d[0]} {d[1]} {d[2]}".format(
                d=doc["spectral_kpoints_mp_grid"]
            )
        )
    elif "spectral_kpoints_list" in doc:
        flines.append("\n%BLOCK SPECTRAL_KPOINTS_LIST")
        for point in doc["spectral_kpoints_list"]:
            flines.append("{p[0]} {p[1]} {p[2]}".format(p=point))
        flines.append("%ENDBLOCK SPECTRAL_KPOINTS_LIST")
    elif "spectral_kpoints_path" in doc:
        flines.append("\n%BLOCK SPECTRAL_KPOINTS_PATH")
        for ind, point in enumerate(doc["spectral_kpoints_path"]):
            flines.append("{p[0]} {p[1]} {p[2]}".format(p=point))
            if "spectral_kpoints_path_labels" in doc:
                flines[-1] += " ! {}".format(doc["spectral_kpoints_path_labels"][ind])
        flines.append("%ENDBLOCK SPECTRAL_KPOINTS_PATH")
    if "spectral_kpoints_path_spacing" in doc:
        flines.append(
            "\nSPECTRAL_KPOINTS_PATH_SPACING: {}".format(
                doc["spectral_kpoints_path_spacing"]
            )
        )

    # specify phonon kpoints
    if "phonon_kpoint_list" in doc:
        flines.append("\n%BLOCK PHONON_KPOINT_LIST")
        for point in doc["phonon_kpoint_list"]:
            flines.append("{p[0]} {p[1]} {p[2]}".format(p=point))
        flines.append("%ENDBLOCK PHONON_KPOINT_LIST")
    elif "phonon_kpoint_mp_spacing" in doc:
        flines.append(
            "\nPHONON_KPOINT_MP_SPACING: {}".format(doc["phonon_kpoint_mp_spacing"])
        )
    elif "phonon_kpoint_mp_grid" in doc:
        flines.append(
            "\nPHONON_KPOINT_MP_GRID: {d[0]} {d[1]} {d[2]}".format(
                d=doc["phonon_kpoint_mp_grid"]
            )
        )

    # specify phonon fine kpoints
    if "phonon_fine_kpoint_list" in doc:
        flines.append("\n%BLOCK PHONON_FINE_KPOINT_LIST")
        for point in doc["phonon_fine_kpoint_list"]:
            flines.append("{p[0]} {p[1]} {p[2]}".format(p=point))
        flines.append("%ENDBLOCK PHONON_FINE_KPOINT_LIST")
    elif "phonon_fine_kpoint_mp_spacing" in doc:
        flines.append(
            "\nPHONON_FINE_KPOINT_MP_SPACING: {}".format(
                doc["phonon_fine_kpoint_mp_spacing"]
            )
        )
    elif "phonon_fine_kpoint_mp_grid" in doc:
        flines.append(
            "\nPHONON_FINE_KPOINT_MP_GRID: {d[0]} {d[1]} {d[2]}".format(
                d=doc["phonon_fine_kpoint_mp_grid"]
            )
        )
    elif "phonon_fine_kpoint_path" in doc:
        flines.append("\n%BLOCK PHONON_FINE_KPOINT_PATH")
        for ind, point in enumerate(doc["phonon_fine_kpoint_path"]):
            flines.append("{p[0]} {p[1]} {p[2]}".format(p=point))
            if "phonon_fine_kpoint_path_labels" in doc:
                flines[-1] += " ! {}".format(doc["phonon_fine_kpoint_path_labels"][ind])
        flines.append("%ENDBLOCK PHONON_FINE_KPOINT_PATH")
    if "phonon_fine_kpoint_path_spacing" in doc:
        flines.append(
            "\nPHONON_FINE_KPOINT_PATH_SPACING: {}".format(
                doc["phonon_fine_kpoint_path_spacing"]
            )
        )

    if "phonon_supercell_matrix" in doc:
        flines.append("\n%BLOCK PHONON_SUPERCELL_MATRIX")
        for i in range(3):
            flines.append(
                "{d[0]:3d} {d[1]:3d} {d[2]:3d}".format(
                    d=doc["phonon_supercell_matrix"][i]
                )
            )
        flines.append("%ENDBLOCK PHONON_SUPERCELL_MATRIX")

    if "cell_constraints" in doc:
        flines.append("\n%BLOCK CELL_CONSTRAINTS")
        flines.append("{d[0]} {d[1]} {d[2]}".format(d=doc["cell_constraints"][0]))
        flines.append("{d[0]} {d[1]} {d[2]}".format(d=doc["cell_constraints"][1]))
        flines.append("%ENDBLOCK CELL_CONSTRAINTS")

    if doc.get("fix_com"):
        flines.append("FIX_COM: TRUE")
    if doc.get("fix_all_cell"):
        flines.append("FIX_ALL_CELL: TRUE")
    if doc.get("fix_all_ions"):
        flines.append("FIX_ALL_IONS: TRUE")
    if doc.get("fix_vol"):
        flines.append("FIX_VOL: TRUE")

    if "symmetry_generate" in doc:
        flines.append("SYMMETRY_GENERATE")
    if "snap_to_symmetry" in doc:
        flines.append("SNAP_TO_SYMMETRY")
    if "symmetry_tol" in doc:
        flines.append("SYMMETRY_TOL: {}".format(doc["symmetry_tol"]))
    if "hubbard_u" in doc:
        flines.append("\n%BLOCK HUBBARD_U")
        flines.append("eV")
        for elem in doc["hubbard_u"]:
            for orbital in doc["hubbard_u"][elem]:
                shift = str(doc["hubbard_u"][elem][orbital])
                flines.append("{} {} {}".format(elem, orbital, shift))
        flines.append("%ENDBLOCK HUBBARD_U")
    if "quantisation_axis" in doc:
        flines.append("QUANTISATION_AXIS: ")
        for integer in doc["quantisation_axis"]:
            flines.append(str(integer) + " ")
    if "positions_noise" in doc:
        flines.append("POSITIONS_NOISE: {}".format(doc["positions_noise"]))
    if "cell_noise" in doc:
        flines.append("CELL_NOISE : {}".format(doc["cell_noise"]))
    if "species_pot" in doc:
        flines.append("\n%BLOCK SPECIES_POT")
        for elem in doc["species_pot"]:
            if elem == "library":
                flines.append(doc["species_pot"]["library"])
            elif "atom_types" not in doc or elem in doc.get("atom_types"):
                flines.append("{:4} {}".format(elem, doc["species_pot"][elem]))
        flines.append("%ENDBLOCK SPECIES_POT")

    ext = "cell"

    return flines, ext


def doc2pdb(doc, path, info=True, hash_dupe=True):
    """Write a simple .pdb for single doc.

    Parameters:
        doc (dict): matador document containing structure.
        path (str): desired path of xsf file.

    Keyword arguments:
        info (bool): write info string to HEADER.
        hash_dupe (bool): hash duplicate file names or skip?

    """
    warnings.warn(
        "This method is no longer maintained and may be removed at any time.",
        DeprecationWarning,
    )
    if path.endswith(".pdb"):
        path = path.replace(".pdb", "")
    if os.path.isfile(path + ".pdb"):
        if hash_dupe:
            path += "-" + generate_hash()
        else:
            raise RuntimeError("Skipping duplicate structure...")
    with open(path + ".pdb", "w") as f:
        try:
            header = "HEADER    {} {}".format(doc["text_id"][0], doc["text_id"][1])
        except Exception:
            header = "HEADER    Generated with matador."
        try:
            # write res file header if info
            title = "TITLE     "
            title += path.split("/")[-1] + " "
            if not doc.get("pressure"):
                title += "0.00 "
            else:
                title += str(doc["pressure"]) + " "
            title += str(doc["cell_volume"]) + " "
            title += str(doc["enthalpy"]) + " "
            title += "0 0 "  # spin
            title += str(doc["num_atoms"]) + " "
            try:
                if "x" in doc["space_group"]:
                    title += "(P1) "
                else:
                    title += "(" + str(doc["space_group"]) + ")" + " "
            except Exception:
                title += "(P1) "
            title += "n - 1"
        except Exception:
            if not info:
                title = "TITLE\t" + path.split("/")[-1]
            raise RuntimeError("Failed to get info for res file, turn info off.")
        author = "AUTHOR    Generated with matador (Matthew Evans, 2016)"
        f.write(header + "\n")
        f.write(author + "\n")
        f.write(title + "\n")
        # use dummy SG for CRYST1, shouldn't matter
        cryst = "CRYST1 {v[0][0]:9.3f} {v[0][1]:9.3f} {v[0][2]:9.3f} {v[1][0]:7.2f} {v[1][1]:7.2f} {v[1][2]:7.2f} P 1".format(
            v=doc["lattice_abc"]
        )
        f.write(cryst + "\n")
        scale_n = cart2abcstar(doc["lattice_cart"])
        f.write(
            "SCALE1    {v[0][0]:10.6f} {v[0][1]:10.6f} {v[0][2]:10.6f}      {:10.5f}\n".format(
                0.0, v=scale_n
            )
        )
        f.write(
            "SCALE2    {v[1][0]:10.6f} {v[1][1]:10.6f} {v[1][2]:10.6f}      {:10.5f}\n".format(
                0.0, v=scale_n
            )
        )
        f.write(
            "SCALE3    {v[2][0]:10.6f} {v[2][1]:10.6f} {v[2][2]:10.6f}      {:10.5f}\n".format(
                0.0, v=scale_n
            )
        )
        if "positions_abs" not in doc:
            doc["positions_abs"] = frac2cart(doc["lattice_cart"], doc["positions_frac"])
        for ind, atom in enumerate(doc["atom_types"]):
            hetatm = "HETATM "
            # append 00 to atom type, a la cell2pdb...
            hetatm += "{:4d} {:.4} NON A   1     ".format(ind + 1, atom + "00")
            hetatm += "{v[0]:7.3f} {v[1]:7.3f} {v[2]:7.3f} {:5.2f} {:5.2f}          {:.2}".format(
                1.0, 0.0, atom, v=doc["positions_abs"][ind]
            )
            f.write(hetatm + "\n")
        ter = "TER       {}       NON A   1".format(len(doc["atom_types"]))
        f.write(ter + "\n")
        f.write("END")


def doc2json(doc, path, overwrite=False, hash_dupe=True):
    """Return raw JSON document as stored in database.

    Parameters:
        doc (dict): matador document containing structure
        path (str): desired filename for res file

    Keyword arguments:
        hash_dupe (bool): hash duplicate file names, or skip?
        overwrite (bool): overwrite if filename exists.

    """
    import json

    if path.endswith(".json"):
        path = path.replace(".json", "")

    if os.path.isfile(path + ".json"):
        if overwrite:
            os.remove(path + ".json")
        elif hash_dupe:
            path += "-" + generate_hash()

    if "_id" in doc:
        doc["_id"] = str(doc["_id"])

    with open(path + ".json", "w") as f:
        f.write(json.dumps(doc, skipkeys=True, indent=2))


def doc2pwscf(doc, path, template=None, spacing=None):
    """Write the structural part of QE input file based
    on the provided matador doc. Will calculate the correct
    kpoint_mp_grid if spacing is provided.

    Parameters:
        doc (dict): matador document containing structure
        path (str): desired filename for res file

    Keyword Arguments:
        template (str): filename of template to prepend before structure
            (with prefix keyword to be replaced),
        spacing (float): kpoint_mp_spacing to use when calculating grid.

    """
    warnings.warn(
        "This method is no longer maintained and may be removed at any time.",
        DeprecationWarning,
    )
    if path.endswith(".in"):
        path = path.replace(".in", "")

    if os.path.isfile(path + ".in"):
        return

    if "lattice_cart" not in doc:
        doc["lattice_cart"] = abc2cart(doc["lattice_abc"])

    if "kpoints_mp_spacing" in doc or spacing is not None:
        if "kpoints_mp_spacing" in doc:
            spacing = doc["kpoints_mp_spacing"]
        doc["kpoints_mp_grid"] = calc_mp_grid(doc["lattice_cart"], spacing)

    file_string = ""
    file_string += "CELL_PARAMETERS angstrom\n"
    for i in range(3):
        file_string += " {d[0]: 10.10f} {d[1]: 10.10f} {d[2]: 10.10f}\n".format(
            d=doc["lattice_cart"][i]
        )

    file_string += "\n ATOMIC_POSITIONS crystal\n"
    for i in range(len(doc["atom_types"])):
        file_string += "{:4} {d[0]: 10.10f} {d[1]: 10.10f} {d[2]: 10.10f}\n".format(
            doc["atom_types"][i], d=doc["positions_frac"][i]
        )
    file_string += "\nK_POINTS automatic\n"
    file_string += "{d[0]} {d[1]} {d[2]} 0 0 0".format(d=doc["kpoints_mp_grid"])

    if template is not None:
        if os.path.isfile(template):
            with open(template, "r") as f:
                template_string = f.readlines()

    with open(path + ".in", "w") as f:
        for line in template_string:
            if "prefix" in line:
                line = "  prefix = '{}',\n".format(path)
            elif "nat" in line:
                line = "  nat = {},\n".format(len(doc["atom_types"]))
            elif "nat" in line:
                line = "  ntyp = {},\n".format(len(set(doc["atom_types"])))
            f.write(line)
        f.write(file_string)


@file_writer_function
def doc2res(
    doc,
    path,
    overwrite=False,
    hash_dupe=False,
    info=True,
    spoof_titl=False,
    sort_atoms=True,
):
    """Write .res file for single doc.

    Parameters:
        doc (dict): matador document containing structure
        path (str): desired filename for res file

    Keyword Arguments:
        info (bool): require info in res file header
        spoof_titl (bool): make up fake info for file header (for use with e.g. cryan)
        sorted (bool): if False, atoms are not sorted (this will not be a valid res file)
        overwrite (bool): whether or not to overwrite colliding files.
        hash_dupe (bool): whether or not to create a unique filename for
            any colliding files, or just skip writing them.

    """
    if spoof_titl:
        info = False
    if not info:
        space_group = "P1" if "space_group" not in doc else doc["space_group"]
        if spoof_titl:
            titl = ("TITL {} -1 1 -1 0 0 {} ({}) n - 1").format(
                path.split("/")[-1], doc["num_atoms"], space_group
            )
        else:
            titl = "TITL file generated by matador (Matthew Evans 2016)"
    else:
        try:
            titl = "TITL "
            titl += path.split("/")[-1] + " "
            if "pressure" not in doc or isinstance(doc["pressure"], str):
                titl += "0.00 "
            else:
                titl += str(doc["pressure"]) + " "
            if "cell_volume" not in doc:
                titl += "0.0 "
            else:
                titl += str(doc["cell_volume"]) + " "
            if "enthalpy" in doc and not isinstance(doc["enthalpy"], str):
                titl += str(doc["enthalpy"]) + " "
            elif "0K_energy" in doc:
                titl += str(doc["0K_energy"]) + " "
            elif "total_energy" in doc:
                titl += str(doc["total_energy"]) + " "
            else:
                raise KeyError("No energy field found.")
            titl += "0 0 "  # spin
            titl += str(doc["num_atoms"]) + " "
            if "space_group" not in doc:
                titl += "(P1) "
            elif "x" in doc["space_group"]:
                titl += "(P1) "
            else:
                titl += "(" + str(doc["space_group"]) + ")" + " "
            titl += "n - 1"
        except Exception:
            raise RuntimeError("Failed to get info for res file, turn info off.")
    if "encapsulated" in doc and doc["encapsulated"]:
        rem = (
            "REM NTPROPS {{'chiralN': {}, 'chiralM': {}, 'r': {}, "
            "'offset': [0.5, 0.5, 0.5], 'date': 'xxx', 'eformperfu': 12345, 'z': {}}}".format(
                doc["cnt_chiral"][0],
                doc["cnt_chiral"][1],
                doc["cnt_radius"],
                doc["cnt_length"],
            )
        )
    flines = []
    if "encapsulated" in doc and doc["encapsulated"]:
        flines.append(rem)
    flines.append(titl)
    cell_str = "CELL 1.0 "
    if (
        "lattice_abc" not in doc
        or len(doc["lattice_abc"]) != 2
        or len(doc["lattice_abc"][0]) != 3
        or len(doc["lattice_abc"][1]) != 3
    ):
        try:
            doc["lattice_abc"] = cart2abc(doc["lattice_cart"])
        except Exception:
            raise RuntimeError(
                "Failed to get lattice, something has gone wrong for {}".format(path)
            )
    for vec in doc["lattice_abc"]:
        for coeff in vec:
            cell_str += "{:.12f} ".format(coeff)
    flines.append(cell_str)
    flines.append("LATT -1")

    # enforce correct order by elements, sorting only the atom_types, not the positions inside them
    if len(doc["positions_frac"]) != len(doc["atom_types"]):
        raise RuntimeError("Atom/position array mismatch!")

    if "site_occupancy" in doc:
        if len(doc["site_occupancy"]) != len(doc["positions_frac"]):
            raise RuntimeError("Occupancy/position array mismatch!")

    if "site_occupancy" not in doc:
        occupancies = [1.0] * len(doc["positions_frac"])

    if sort_atoms:
        positions_frac, atom_types = zip(
            *[
                (pos, types)
                for (types, pos) in sorted(
                    zip(doc["atom_types"], doc["positions_frac"]), key=lambda k: k[0]
                )
            ]
        )
        if "site_occupancy" in doc:
            occupancies, _atom_types = zip(
                *[
                    (occ, types)
                    for (types, occ) in sorted(
                        zip(doc["atom_types"], doc["site_occupancy"]),
                        key=lambda k: k[0],
                    )
                ]
            )

    else:
        positions_frac = doc["positions_frac"]
        atom_types = doc["atom_types"]

    if len(positions_frac) != len(doc["positions_frac"]) or len(atom_types) != len(
        doc["atom_types"]
    ):
        raise RuntimeError("Site occupancy mismatch!")

    written_atoms = []
    sfac_str = "SFAC \t"
    for elem in atom_types:
        if elem not in written_atoms:
            sfac_str += " " + str(elem)
            written_atoms.append(str(elem))
    flines.append(sfac_str)
    atom_labels = []
    i = 0
    j = 1
    while i < len(atom_types):
        num = atom_types.count(atom_types[i])
        atom_labels.extend(num * [j])
        i += num
        j += 1

    for atom in zip(atom_types, atom_labels, positions_frac, occupancies):
        flines.append(
            "{0:8s}{1:3d}{2[0]: 20.15f} {2[1]: 20.15f} {2[2]: 20.15f}  {3: 20.15f}".format(
                atom[0], atom[1], atom[2], atom[3]
            )
        )
    flines.append("END")
    # very important newline for compatibliy with cryan
    # flines.append('')

    return flines, "res"


def doc2xsf(doc, path, write_energy=False, write_forces=False, overwrite=False):
    """Write an .xsf file for a matador document, with positions in
    Cartesian coordinates. Optionally, write the energy in a comment
    at the top of the file for use with aenet.

    Parameters:
        doc (dict): matador document containing structure.
        path (str): desired path of xsf file.

    Keyword arguments:
        write_energy (bool): whether or not to write total energy in a comment
            as the first line of the file.
        write_forces (bool): whether or not to write the forces on each atom.
        overwrite (bool): overwrite if file exists.

    """
    warnings.warn(
        "This method is no longer maintained and may be removed at any time.",
        DeprecationWarning,
    )
    if path.endswith(".xsf"):
        path = path.replace(".xsf", "")

    flines = []
    if write_energy:
        if "total_energy" in doc:
            flines.append("# total energy = {:10.8f} eV\n".format(doc["total_energy"]))
        else:
            raise RuntimeError(
                "Failed to write energy in xsf file: key 'total_energy' missing from input."
            )

    flines.append("CRYSTAL")
    flines.append("PRIMVEC")
    if "lattice_cart" in doc:
        for i in range(3):
            flines.append(
                "\t\t{lat[0]: 10.8f}\t{lat[1]: 10.8f}\t{lat[2]: 10.8f}".format(
                    lat=doc["lattice_cart"][i]
                )
            )
    else:
        raise RuntimeError(
            "Failed to write lattice in xsf file: key 'lattice_cart' missing from input."
        )
    flines.append("PRIMCOORD")
    flines.append("{} {}".format(doc["num_atoms"], 1))
    if "positions_abs" not in doc:
        doc["positions_abs"] = frac2cart(doc["lattice_cart"], doc["positions_frac"])
    for ind, (atom, position) in enumerate(
        zip(doc["atom_types"], doc["positions_abs"])
    ):
        flines.append(
            "{:2}\t{pos[0]: 16.8f}\t{pos[1]: 16.8f}\t{pos[2]: 16.8f}".format(
                atom, pos=position
            )
        )
        if write_forces:
            if "forces" in doc:
                flines[-1] += "\t{f[0]: 16.8f}\t{f[1]: 16.8f}\t{f[2]: 16.8f}".format(
                    f=doc["forces"][ind]
                )
            else:
                raise RuntimeError(
                    "Failed to write forces in xsf file: key 'forces' missing from input."
                )

    if os.path.isfile(path + ".xsf"):
        if overwrite:
            os.remove(path + ".xsf")
        else:
            raise RuntimeError("Duplicate file!")

    with open(path + ".xsf", "w") as f:
        for line in flines:
            f.write(line + "\n")


@file_writer_function
def doc2arbitrary(doc, path, overwrite=False, hash_dupe=False):
    """Write a Python dictionary into a standard CASTEP-style
    keyword: value file.

    Parameters:
        doc (dict): contains key-value pairs.
        path (str): filename to write to.

    Keyword arguments:
        hash_dupe (bool): hash duplicate file names, or skip?
        overwrite (bool): overwrite if filename exists.

    Returns:
        list: list of strs to write to file.

    """
    warnings.warn(
        "This method is no longer maintained and may be removed at any time.",
        DeprecationWarning,
    )
    ext = None
    flines = []
    output_doc = {}
    # sanitise dict to include only unique-by-case keys
    for key in doc:
        if key.startswith("_"):
            continue
        if key.lower() == "source":
            continue
        if key.lower() in output_doc:
            raise Warning("Key {} defined multiple times.".format(key))
        output_doc[key.lower()] = doc[key]

    for key in output_doc:
        flines.append("{}: {}".format(key, output_doc[key]))
    return flines, ext



# coding: utf-8
# Distributed under the terms of the MIT License.

""" Run calculations in a folder with BatchRun
such that there are no clashes.
"""
import os
import argparse
from matador import __version__, script_epilog
from matador.utils.print_utils import print_notify
from matador.utils.errors import InputError
from matador.compute import BatchRun


def main():
    """Parse args and run any remaining jobs."""
    parser = argparse.ArgumentParser(
        prog="run3",
        description='Run multiple calculations from a series of .res \
                     files and single cell and param files, typically CASTEP geometry \
                     optimisations. The geometry optimization will \
                     be split into "--rough" (default: 4) chunks of "--rough_iter" (default: 2) \
                     iterations, followed by chunks of "--fine_iter" (default: 20) \
                     iterations, until geom_max_iter is reached in the param file. \
                     Successful runs will be moved to "completed", crashes/failures will go to \
                     "bad_castep" and initial res files will go into "input". Running jobs will \
                     be listed in jobs.txt and those that completed cleanly will be listed \
                     in finished_cleanly.txt.',
        epilog=script_epilog,
    )
    parser.add_argument(
        "--version", action="version", version="matador version " + __version__ + "."
    )
    parser.add_argument(
        "seed",
        type=str,
        nargs="+",
        help="cell and param seed to use as template for CASTEP calculations OR list of files\
                              to apply run executable on",
    )
    parser.add_argument(
        "-nc",
        "--ncores",
        type=int,
        help="number of cores per node per job [DEFAULT=cpu_count/nprocesses]",
    )
    parser.add_argument(
        "-np",
        "--nprocesses",
        type=int,
        default=1,
        help="number of concurrent calculations, i.e. number \
                              of concurrent mpiruns [DEFAULT=1]",
    )
    parser.add_argument(
        "-nn",
        "--nnodes",
        type=int,
        default=1,
        help="number of nodes per job, i.e. number of nodes \
                              using -nc cores [DEFAULT=1].",
    )

    parser.add_argument(
        "-t",
        "--max_walltime",
        type=int,
        help="maximum walltime in seconds (job will quit early to clean up if specified)",
    )
    parser.add_argument(
        "-exec",
        "--executable",
        type=str,
        help="specify path to or name of executable (DEFAULT: castep)",
    )
    parser.add_argument(
        "--no_reopt",
        action="store_true",
        default=False,
        help="do not run geometry optimisation again after first success",
    )
    parser.add_argument(
        "--redirect",
        type=str,
        help="filename to redirect output to, can use $seed macro",
    )
    parser.add_argument(
        "--mode", type=str, default="castep", help="either castep or generic"
    )
    parser.add_argument(
        "--noise",
        action="store_true",
        help=(
            "add 0.1 A of random noise to positions on every cell, "
            "useful for converging forces (DEFAULT: off)"
        ),
    )
    parser.add_argument(
        "--squeeze",
        action="store_true",
        help="add external pressure to the rough steps of geom opts",
    )
    parser.add_argument(
        "--ignore_jobs_file",
        action="store_true",
        help="whether to use the jobs.txt file to avoid clashes",
    )
    parser.add_argument(
        "-d", "--debug", action="store_true", default=False, help="debug output"
    )
    parser.add_argument(
        "-cust",
        "--custom_params",
        action="store_true",
        default=False,
        help="use custom param file per structure",
    )
    parser.add_argument(
        "-v",
        "--verbosity",
        type=int,
        default=2,
        help="integer to set level of verbosity",
    )
    parser.add_argument(
        "--archer", action="store_true", default=False, help="use aprun over mpirun"
    )
    parser.add_argument(
        "--slurm", action="store_true", default=False, help="use srun over mpirun"
    )
    parser.add_argument(
        "--intel", action="store_true", default=False, help="use Intel's mpirun"
    )
    parser.add_argument(
        "--conv_cutoff",
        action="store_true",
        default=False,
        help="run all res files at cutoff defined in cutoff.conv file",
    )
    parser.add_argument(
        "--conv_kpt",
        action="store_true",
        default=False,
        help="run all res files at kpoint spacings defined in kpt.conv file",
    )
    parser.add_argument(
        "--memcheck",
        action="store_true",
        default=False,
        help="enable memcheck via castep dryrun",
    )
    parser.add_argument(
        "--scratch_prefix",
        type=str,
        help="specify absolute path prefix for compute dir e.g. "
        "--scratch_prefix /scratch/user/ will set the compute directory to /scratch/user/$hostname. "
        "default value is taken from .matadorrc.",
    )
    parser.add_argument("--maxmem", type=int, help="override max memory for memcheck")
    parser.add_argument(
        "--killcheck",
        action="store_true",
        default=True,
        help="check for $seed.kill file and quit job if present",
    )
    parser.add_argument(
        "--kpts_1D",
        action="store_true",
        default=False,
        help="recalculate a 1D kpoint mesh of spacing specified in template cell",
    )
    parser.add_argument(
        "--spin",
        type=int,
        nargs="?",
        const=5,
        default=None,
        help=(
            "if not specified in .cell file, break spin symmetry on first atom using the spin specified by "
            "the user [DEFAULT: 5]"
        ),
    )
    parser.add_argument(
        "--rough",
        type=int,
        default=4,
        help="choose how many <rough_iter> geometry optimizations \
                              to perform, decrease if lattice is nearly correct. [DEFAULT: 4].",
    )
    parser.add_argument(
        "--rough_iter",
        type=int,
        default=2,
        help="choose how many relaxation steps per rough geometry optimization\
                              to perform, [DEFAULT: 2].",
    )
    parser.add_argument(
        "--fine_iter",
        type=int,
        default=20,
        help="choose how many relaxation steps per fine geometry optimization\
                              to perform, [DEFAULT: 20].",
    )
    parser.add_argument(
        "-l", "--limit", type=int, default=None, help="limit to n structures per run"
    )
    parser.add_argument(
        "--profile", action="store_true", help="profile code with cProfile"
    )
    args = parser.parse_args()

    seed = vars(args)["seed"]

    kwargs = vars(args)
    del kwargs["seed"]

    from matador.config import load_custom_settings

    settings = load_custom_settings(debug=kwargs.get("debug")).get("run3")
    if settings is not None:
        if kwargs["scratch_prefix"] is None:
            kwargs["scratch_prefix"] = settings.get("scratch_prefix")
            if kwargs["scratch_prefix"] == ".":
                kwargs["scratch_prefix"] = None

        if kwargs["executable"] is None:
            kwargs["executable"] = settings.get("castep_executable", "castep")

        kwargs["run3_settings"] = settings

    if sum([vars(args)["slurm"], vars(args)["archer"], vars(args)["intel"]]) > 1:
        exit(
            "Incompatible MPI arguments specified, please use at most one of --archer/--intel/--slurm."
        )

    if vars(args).get("profile"):
        import cProfile
        import pstats
        from sys import version_info

        hostname = os.uname()[1]
        pr = cProfile.Profile()
        pr.enable()

    try:
        runner = BatchRun(seed, **kwargs)
        runner.spawn()
    except InputError as exc:
        print_notify(exc)
    except RuntimeError:
        print_notify("Some jobs failed, exiting...")
    except Exception as exc:
        raise exc

    if vars(args).get("profile"):
        pr.disable()
        fname = "run3-{}-{}-{}.{}.{}".format(
            __version__,
            hostname,
            version_info.major,
            version_info.minor,
            version_info.micro,
        )
        pr.dump_stats(fname + ".prof")
        with open(fname + ".pstats", "w") as s:
            sortby = "cumulative"
            ps = pstats.Stats(pr, stream=s).sort_stats(sortby)
            ps.print_stats()


if __name__ == "__main__":
    main()
